@@ -1,30 +1,30 @@
 /*
  * f p o r t . c				-- File ports
- * 
- * Copyright  2000-2001 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
- * 
- * 
+ *
+ * Copyright  2000-2003 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
+ *
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
- * 
+ *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
+ * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
  * USA.
- * 
+ *
  *           Author: Erick Gallesio [eg@unice.fr]
  *    Creation date:  8-Jan-2000 14:48 (eg)
- * Last file update: 16-Jan-2001 23:19 (eg)
+ * Last file update: 10-Feb-2003 11:15 (eg)
  *
  * This implementation is built by reverse engineering on an old SUNOS 4.1.1
- * stdio.h. It has been simplified to fit the needs for STklos. In particular 
+ * stdio.h. It has been simplified to fit the needs for STklos. In particular
  * non buffered file are not implemented. Anyway this is faster than an
  * implementation using the C buffered IO (at least on glibc)
  *
@@ -32,13 +32,24 @@
 #include "stklos.h"
 
 
-char *STk_current_filename;		  /* Name of the file we read */ 
+char *STk_current_filename;		  /* Name of the file we read */
 int STk_line_counter, STk_column_counter; /* Position when reading a file */
 
 SCM STk_curr_iport, STk_curr_oport, STk_curr_eport;  /* current active ports   */
 SCM STk_stdin, STk_stdout, STk_stderr;		     /* The unredirected ports */
 
 
+/*
+ * Since we manage ourselves our port, we need to kee a reference on all
+ * the open ports such that all the port are flushed when we exit from
+ * the program.
+ * FIXME: This method is not optimal. What we should have is a finalizer on
+ * ports (that is what we have til 0.51 version) AND a weak list of the open
+ * ports. With a weak list port can be closed on finalization, whereas in our
+ * solution, unclosed port which can be garbaged will not be freed because of 
+ * the reference through the global all_fports
+ */
+static SCM all_fports = STk_nil;
 
 /*===========================================================================*\
  * 
@@ -68,7 +79,7 @@ struct fstream {
   char *filename;
   SCM  read_event;
   SCM  write_event;
-  SCM  idle_proc;
+  SCM  idle_procs;
 };
 
 
@@ -82,7 +93,7 @@ struct fstream {
 #define PORT_NAME(x)		(((struct fstream *) (x))->filename)
 #define PORT_REVENT(x)		(((struct fstream *) (x))->read_event)
 #define PORT_WEVENT(x)		(((struct fstream *) (x))->write_event)
-#define PORT_IDLE(x)		(((struct fstream *) (x))->idle_proc)
+#define PORT_IDLE(x)		(((struct fstream *) (x))->idle_procs)
 
 
 /*===========================================================================*\
@@ -166,19 +177,20 @@ static inline int Freadyp(void *stream)
 
     FD_ZERO(&readfds);
     FD_SET(f, &readfds);
-    timeout.tv_sec = 0; timeout.tv_usec = 0; /* //FIXME: 0=>charge trop importante*/
+    timeout.tv_sec = 0; timeout.tv_usec = 10; /* FIXME: 0=>charge trop importante*/
     return (select(f+1, &readfds, NULL, NULL, &timeout));
   }
 }
 
 static inline int Fgetc(void *stream)
 {
-  if (PORT_IDLE(stream) != STk_false) {
-    /* There is an idle handler */
-    SCM idle = PORT_IDLE(stream);
+  if (PORT_IDLE(stream) != STk_nil) {
+    /* There is at least an idle handler */
+    SCM idle;
     
     while (!Freadyp(stream))
-      STk_C_apply(idle, 0);
+      for (idle = PORT_IDLE(stream); !NULLP(idle); idle = CDR(idle))
+	STk_C_apply(CAR(idle), 0);
   }
 
   for ( ; ; ) {
@@ -200,7 +212,7 @@ static inline int Fread(void *stream, void *buff, int count)
 static int Fclose(void *stream)		/* File version */
 {
   int ret = flush_buffer(stream);
-  
+
   return (ret == EOF) ? EOF : close(PORT_FD(stream));
 }
 
@@ -270,12 +282,20 @@ static void fport_print(SCM obj, SCM port) 	/* Generic printing of file ports */
 }
 
 
+static void fport_release(SCM port)	/* remove port from list of open ports */
+{
+  all_fports = STk_dremq(port, all_fports);
+}
+
+
+
+#ifdef STK_PORT_FINALIZER
 static void fport_finalizer(struct port_obj *port)
 {
   /* Close the associated stream */
   STk_close((SCM) port);
 }
-
+#endif
 
 static struct port_obj *
 make_fport(char *filename, FILE *f, int flags)
@@ -306,7 +326,7 @@ make_fport(char *filename, FILE *f, int flags)
   PORT_NAME(fs)		 = STk_strdup(filename);
   PORT_REVENT(fs)	 = STk_false;
   PORT_WEVENT(fs)	 = STk_false;
-  PORT_IDLE(fs)		 = STk_false;
+  PORT_IDLE(fs)		 = STk_nil;
 
   /* Initialize now the port itsef */
   NEWCELL(res, port);
@@ -318,6 +338,7 @@ make_fport(char *filename, FILE *f, int flags)
   PORT_POS(res)		= 0;
 
   PORT_PRINT(res)	= fport_print;
+  PORT_RELEASE(res)	= fport_release;
   PORT_GETC(res)	= Fgetc;
   PORT_READY(res)	= Freadyp;
   PORT_EOFP(res)	= Feof;
@@ -328,13 +349,18 @@ make_fport(char *filename, FILE *f, int flags)
   PORT_BREAD(res)	= Fread;
   PORT_BWRITE(res)	= Fwrite;
 
+#ifdef STK_PORT_FINALIZER
   /* Add a finalizer on file to close it when the GC frees it */
   STk_register_finalizer(res, fport_finalizer);
+#endif
+  /* Add res to the list of open ports */ 
+  all_fports = STk_cons(res, all_fports);
+  
   return (struct port_obj *)res;
 }
 
 
-#ifdef __CYGWIN32__
+#ifdef WIN32
 static char *convert_for_win32(char *mode)
 {
   /* Things are complicated on Win32 (as always). So we onvert all files 
@@ -369,9 +395,9 @@ static SCM open_file_port(SCM filename, char *mode, int flags, int error)
   name = STRING_CHARS(filename);
 
   if (strncmp(name, "| ", 2)) {
-    full_name  = STRING_CHARS(STk_int_expand_file_name(name));
+    full_name  = STk_expand_file_name(name);
     flags     |= PORT_IS_FILE;
-#ifdef __CYGWIN32__
+#ifdef WIN32
     mode = convert_for_win32(mode);
 #endif
     if ((f = fopen(full_name, mode)) == NULL) {
@@ -567,17 +593,44 @@ DEFINE_PRIMITIVE("rewind-file-port", scheme_rewind_file_port, subr1, (SCM port))
 }
 
 
-
-DEFINE_PRIMITIVE("%port-idle-set!", port_idle_set, subr2, (SCM port, SCM proc))
+/*
+<doc EXT port-file-name
+ * (port-file-name port)
+ *
+ * Returns the file name used to open |port|; |port| must be a file port. 
+doc>
+ */
+DEFINE_PRIMITIVE("port-file-name", port_file_name, subr1, (SCM port))
 {
-  ENTER_PRIMITIVE(scheme_rewind_file_port);
-  if (!FPORTP(port)) 			 error_bad_file_port(port);
-  if (STk_procedurep(proc) == STk_false) STk_error("bad procedure ~S", proc);
-
-  PORT_IDLE(PORT_STREAM(port)) = proc;
-  return STk_void;
+  ENTER_PRIMITIVE(port_file_name);
+  if (!FPORTP(port)) error_bad_file_port(port);
+  return STk_Cstring2string(PORT_NAME(PORT_STREAM(port)));
 }
 
+
+DEFINE_PRIMITIVE("%port-file-fd", port_file_fd, subr1, (SCM port))
+{
+  ENTER_PRIMITIVE(port_file_fd);
+  return FPORTP(port)? MAKE_INT(PORT_FD(PORT_STREAM(port))) : STk_false;
+}
+
+
+DEFINE_PRIMITIVE("%port-idle", port_idle, subr12, (SCM port, SCM val))
+{
+  ENTER_PRIMITIVE(port_idle);
+  if (!FPORTP(port)) error_bad_file_port(port);
+
+  if (val) {
+    /* Set the idle list to the given value. No control on the content of 
+     * the procedure list (must be done in Scheme)
+     */
+    if (STk_int_length(val) < 0) STk_error("bad list ~S", val);
+    PORT_IDLE(PORT_STREAM(port)) = val;
+  }
+  return  PORT_IDLE(PORT_STREAM(port));
+}
+
+
 /*===========================================================================*\
  * 
  * Utilities on file ports
@@ -600,18 +653,47 @@ Error:
   return STk_void; /* for the compiler */
 }
 
-SCM STk_set_port_idle(SCM port, SCM idle_func)
+SCM STk_add_port_idle(SCM port, SCM idle_func)
 {
-  PORT_IDLE(PORT_STREAM(port)) = idle_func;
+  PORT_IDLE(PORT_STREAM(port)) = STk_cons(idle_func, 
+					  PORT_IDLE(PORT_STREAM(port)));
   return STk_void;
 }
 
+void STk_set_line_buffered_mode(SCM port)
+{
+  struct fstream* fs;
+  int flags;
+
+  /* Assert: port is a valid input port */
+  fs    = PORT_STREAM(port);
+  flags = PORT_STREAM_FLAGS(fs);
+
+  flags &= ~(STK_IOFBF|STK_IONBF);
+  PORT_STREAM_FLAGS(fs) = flags | STK_IOLBF;
+}
+
+
+/*
+ * Closing and flushing all open port before exiting
+ */
+static void at_exit_close_all_fports(void)
+{
+  SCM l = STk_copy_tree(all_fports); /* Use l since all_fports will be altered */
+
+  while (!NULLP(l)) {
+    /* Don't close the input, since it interact badly with interactive REPL */
+    if (CAR(l) != STk_curr_iport && CAR(l) != STk_stdin)
+      STk_close(CAR(l));
+    l = CDR(l);
+  }
+}
 
 
 int STk_init_fport(void)
 {
   STk_stdin  = STk_curr_iport = (SCM) make_fport("*stdin*",  stdin,
-						 PORT_IS_FILE | PORT_READ);  
+						 PORT_IS_FILE | PORT_READ); 
   STk_stdout = STk_curr_oport = (SCM) make_fport("*stdout*", stdout,
 						 PORT_IS_FILE | PORT_WRITE);
   STk_stderr = STk_curr_eport = (SCM) make_fport("*stderr*", stderr,
@@ -624,9 +706,14 @@ int STk_init_fport(void)
   ADD_PRIMITIVE(output_fportp);
   ADD_PRIMITIVE(scheme_open_file);
   ADD_PRIMITIVE(scheme_rewind_file_port);
+  ADD_PRIMITIVE(port_file_name);
 
-  ADD_PRIMITIVE(port_idle_set);
+  ADD_PRIMITIVE(port_file_fd);
+  ADD_PRIMITIVE(port_idle);
   STk_current_filename	= "";	/* "" <=> stdin */
+
+
+  atexit(at_exit_close_all_fports);
   return TRUE;
 }
 /*  LocalWords:  filename
