@@ -2,7 +2,7 @@
  *
  * h a s h  . c			-- Hash Tables 
  *
- * Copyright  1994-2001 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
+ * Copyright  1994-2002 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
  * 
  +=============================================================================
  ! This code is a rewriting of the file tclHash.c of the Tcl
@@ -36,7 +36,7 @@
  *
  *           Author: Erick Gallesio [eg@kaolin.unice.fr]
  *    Creation date: 17-Jan-1994 17:49
- * Last file update: 16-Feb-2001 12:23 (eg)
+ * Last file update: 24-Apr-2002 12:18 (eg)
  */
 
 #include "stklos.h"
@@ -116,7 +116,7 @@ unsigned long sxhash(SCM obj)
   SCM tmp;
   int i;
   
-  if (!BOXED_OBJP(obj)) return (unsigned long) (obj >> 2);
+  if (!BOXED_OBJP(obj)) return (AS_LONG(obj) >> 2);
 
   switch (BOXED_TYPE(obj)) {
     case tc_cons:       h = sxhash(CAR(obj));
@@ -374,9 +374,10 @@ void STk_hash_set_variable(struct hash_table_obj *h, SCM v, SCM value)
  *
 \*===========================================================================*/
 
+/* Return the list of keys of a hash table */
 SCM STk_hash_keys(struct hash_table_obj *h)
 {
-  int i, max; 
+  int i, max;
   SCM tmp, res = STk_nil;
 
   for (i=0, max= HASH_NBUCKETS(h); i < max; i++) {
@@ -491,7 +492,7 @@ DEFINE_PRIMITIVE("hash-table-put!", hash_put, subr3, (SCM ht, SCM key, SCM val))
 
       for(l = HASH_BUCKETS(ht)[index]; !NULLP(l); l = CDR(l)) 
 	if (STk_C_apply(func, 2, CAR(CAR(l)), key) != STk_false) break;
-    default: 
+    default:  break;
   }
 
   /* When we are here, index is on the good bucket list. if l is null, 
@@ -564,7 +565,7 @@ DEFINE_PRIMITIVE("hash-table-get", hash_get, subr23, (SCM ht, SCM key, SCM def))
       for(l = HASH_BUCKETS(ht)[index]; !NULLP(l); l = CDR(l)) 
 	if (STk_C_apply(func, 2, CAR(CAR(l)), key) != STk_false) return CDR(CAR(l));
       break;
-    default: 
+    default: break;
   }
 
   /* Not found */
@@ -622,7 +623,7 @@ DEFINE_PRIMITIVE("hash-table-remove!", hash_remove, subr2, (SCM ht, SCM key))
       for(l = HASH_BUCKETS(ht)[index]; !NULLP(l); prev=l, l=CDR(l)) 
 	if (STk_C_apply(func, 2, CAR(CAR(l)), key) != STk_false) break;
       break;
-    default: 
+    default: break;
   }
 
   if (!NULLP(l)) {     			/* found item */
