@@ -2,7 +2,7 @@
  *
  * l i s t . c			-- Lists procedures
  *
- * Copyright  1993-2000 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
+ * Copyright  1993-2002 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
  *
  *
  * This program is free software; you can redistribute it and/or modify
@@ -22,7 +22,7 @@
  *
  *           Author: Erick Gallesio [eg@kaolin.unice.fr]
  *    Creation date: ??-Oct-1993 21:37
- * Last file update:  9-Dec-2000 12:11 (eg)
+ * Last file update: 19-Dec-2002 11:12 (eg)
  */
 
 #include "stklos.h"
@@ -77,7 +77,7 @@ SCM STk_argv2list(int argc, SCM *argv)
   SCM res = STk_nil;
   
   while (argc--) {
-    res = STk_cons(argv[argc], res);
+    res = STk_cons(argv[-argc], res);
   }
   return res;
 }
@@ -268,7 +268,7 @@ doc>
 {
   register SCM *tmp, l = STk_nil;
 
-  for (tmp = &argv[argc-1]; tmp >= argv; tmp--)
+  for (tmp = argv-argc+1; tmp <= argv; tmp++)
     l = STk_cons(*tmp, l);
 
   return l;
@@ -357,8 +357,8 @@ DEFINE_PRIMITIVE("append", append, vsubr, (int argc, SCM* argv))
   switch (argc) {
     case 0:  return STk_nil;
     case 1:  return *argv;
-    case 2:  return STk_append2(argv[0], argv[1]);
-    default: return STk_append2(argv[0], STk_append(argc-1, argv+1));
+    case 2:  return STk_append2(argv[0], argv[-1]);
+    default: return STk_append2(argv[0], STk_append(argc-1, argv-1));
   }
 }
 
@@ -578,13 +578,13 @@ Error:								\
 }
 
 DEFINE_PRIMITIVE("assq", assq, subr2, (SCM obj, SCM alist))
-     LASSOC(PTR_EQ, assq);
+     LASSOC(PTR_EQ, assq)
 
 DEFINE_PRIMITIVE("assv", assv, subr2, (SCM obj, SCM alist))
-     LASSOC(PTR_EQV, assv);
+     LASSOC(PTR_EQV, assv)
 
 DEFINE_PRIMITIVE("assoc", assoc, subr2, (SCM obj, SCM alist))
-     LASSOC(PTR_EQUAL, assoc);
+     LASSOC(PTR_EQUAL, assoc)
 
 
 /***
@@ -594,6 +594,24 @@ DEFINE_PRIMITIVE("assoc", assoc, subr2, (SCM obj, SCM alist))
  ***/
 
 
+/*
+<doc EXT pair-mutable?
+ * (pair-mutable? obj)
+ * 
+ * Returns |#t| if |obj| is a mutable pair, otherwise returns |#f|. 
+ * @lisp
+ * (pair-mutable? '(1 . 2))    => #f
+ * (pair-mutable? (cons 1 2))  => #t
+ * (pair-mutable? 12)          => #f
+ * @end lisp
+doc>
+*/
+DEFINE_PRIMITIVE("pair-mutable?", pair_mutable, subr1, (SCM obj))
+{
+  return MAKE_BOOLEAN(CONSP(obj) && !(BOXED_INFO(obj) & CONS_CONST));
+}
+
+
 /*
 <doc EXT list*
  * (list* obj ...)
@@ -612,9 +630,11 @@ DEFINE_PRIMITIVE("list*", list_star, vsubr, (int argc, SCM *argv))
   register SCM *tmp, l;
 
   if (argc == 0) return STk_nil;
-  l = argv[argc-1];
 
-  for (tmp = &argv[argc-2]; tmp >= argv; tmp--)
+  tmp = argv-argc+1;
+  l   = *tmp;
+
+  for (++tmp; tmp <= argv; tmp++)
     l = STk_cons(*tmp, l);
 
   return l;
@@ -753,7 +773,7 @@ DEFINE_PRIMITIVE("filter!", dfilter, subr2, (SCM pred, SCM list))
 doc>
 */
 
-static SCM dappend2(SCM l1, SCM l2)
+SCM STk_dappend2(SCM l1, SCM l2)
 {
   register SCM tmp;
 
@@ -777,8 +797,8 @@ DEFINE_PRIMITIVE("append!", dappend, vsubr, (int argc, SCM* argv))
   switch (argc) {
     case 0:  return STk_nil;
     case 1:  return *argv;
-    case 2:  return dappend2(argv[0], argv[1]);
-    default: return dappend2(argv[0], STk_dappend(argc-1, argv+1));
+    case 2:  return STk_dappend2(argv[0], argv[-1]);
+    default: return STk_dappend2(argv[0], STk_dappend(argc-1, argv-1));
   }
 }
 
@@ -812,6 +832,31 @@ DEFINE_PRIMITIVE("reverse!", dreverse, subr1, (SCM l))
   return prev;
 }
 
+/*
+ *
+ * Utility procedure equivalent to (delete! obj list eq?)
+ *
+ */
+SCM STk_dremq(SCM obj, SCM list)
+{
+  SCM previous, l, start=list;
+
+  for (previous=STk_nil, l=list; !NULLP(l); ) {
+    if (!CONSP(l)) error_bad_list(list);
+    if (BOXED_INFO(l) & CONS_CONST) error_const_cell(l);
+
+    if (obj == CAR(l)) {
+      if (previous == STk_nil)
+	list = CDR(list);
+      else 
+	CDR(previous) = CDR(l);
+    } else {
+      previous = l;
+    }
+    if ((l=CDR(l)) == start) error_circular_list(list);
+  }
+  return list;
+}
 
 
 int STk_init_list(void)
@@ -837,6 +882,7 @@ int STk_init_list(void)
   ADD_PRIMITIVE(assv);
   ADD_PRIMITIVE(assoc);
 
+  ADD_PRIMITIVE(pair_mutable);
   ADD_PRIMITIVE(list_star);
   ADD_PRIMITIVE(copy_tree);
   ADD_PRIMITIVE(last_pair);
