@@ -21,7 +21,7 @@
  *
  *           Author: Erick Gallesio [eg@kaolin.unice.fr]
  *    Creation date: 12-May-1993 10:34
- * Last file update: 19-Feb-2001 23:14 (eg)
+ * Last file update:  7-Feb-2003 21:31 (eg)
  */
 
 
@@ -101,8 +101,8 @@ static SCM my_sqrt(SCM z);
 static SCM my_magnitude(SCM z);
 static SCM my_angle(SCM z);
 
-EXTERN_PRIMITIVE("make-rectangular", make_rectangular, subr2, (SCM r, SCM i));
-EXTERN_PRIMITIVE("real-part", real_part, subr1, (SCM z));
+EXTERN_PRIMITIVE("make-rectangular", make_rectangular, subr2, (SCM r, SCM i))
+EXTERN_PRIMITIVE("real-part", real_part, subr1, (SCM z))
 
 
 /******************************************************************************
@@ -180,7 +180,7 @@ static Inline SCM Cmake_rational(SCM n, SCM d)
 }
 
 
-static int make_rational(SCM n, SCM d)
+static SCM make_rational(SCM n, SCM d)
 {
   SCM gcd;
 
@@ -468,7 +468,7 @@ static char *number2Cstr(SCM n, long base, char buffer[])
 	s2 = number2Cstr(RATIONAL_DEN(n), base, tmp);
 	s3 = STk_must_malloc(strlen(s1) + strlen(s2) + 2);
 	sprintf(s3, "%s/%s", s1, s2);
-	if (s2!=tmp) STk_free(tmp); /*buffer will event. be deallocated by caller*/
+	if (s2!=tmp) STk_free(s2); /*buffer will event. be deallocated by caller*/
 	return s3;
       }
     case tc_complex:
@@ -479,7 +479,7 @@ static char *number2Cstr(SCM n, long base, char buffer[])
 	s2 = number2Cstr(COMPLEX_IMAG(n), base, tmp);
 	s3 = STk_must_malloc(strlen(s1) + strlen(s2) + 3);
 	sprintf(s3, "%s%s%si", s1, ((*s2 == '-') ? "": "+"), s2);
-	if (s2!=tmp) STk_free(tmp); /*buffer will event. be deallocated by caller*/
+	if (s2!=tmp) STk_free(s2); /*buffer will event. be deallocated by caller*/
 	return s3;
       }
     case tc_real:
@@ -601,6 +601,68 @@ SCM STk_ulong2integer(unsigned long n)
    }
 }
 
+#define INT32_MAX   2147483647
+#define UINT32_MAX  4294967295U
+
+
+long STk_integer2int32(SCM n, int *overflow)
+{
+  *overflow = 0;
+
+  if (INTP(n)) {
+#if (LONG_MAX == INT32_MAX)  	  /* longs are on 32 bits */
+    return INT_VAL(n);
+#else				  /* longs are more than 32 bits (probably 64) */
+    long val = INT_VAL(n);
+    
+    if ((- INT32_MAX - 1) <= val && val < INT32_MAX)
+      return val;
+    else {
+      *overflow = 1;
+      return 0;
+    }
+#endif
+  }
+  if (BIGNUMP(n)) {
+    mpz_t *v = &BIGNUM_VAL(n);
+    if (mpz_cmp_si(*v, (- INT32_MAX - 1)) >= 0 && mpz_cmp_si(*v, INT32_MAX) <= 0)
+      return mpz_get_si(*v);
+  }
+  *overflow = 1;
+  return 0;
+}
+
+
+unsigned long STk_integer2uint32(SCM n, int *overflow)
+{
+  *overflow = 0;
+  
+  if (INTP(n)) {
+    long val = INT_VAL(n);
+    
+    if (val >= 0) {
+#if (ULONG_MAX == UINT32_MAX)  	/* unsigned longs are on 32 bits */
+      return (unsigned long) INT_VAL(n);
+#else				/* longs are more than 32 bits (probably 64) */
+      if (val < UINT32_MAX)
+	return val;
+      else {
+	*overflow = 1;
+	return 0;
+      }
+#endif
+    }
+  }
+  if (BIGNUMP(n)) {
+    mpz_t *v = &BIGNUM_VAL(n);
+    if (mpz_cmp_ui(*v, 0) >= 0 && mpz_cmp_ui(*v, UINT32_MAX) <= 0)
+      return mpz_get_ui(*v);
+  }
+  *overflow = 1;
+  return 0;
+}
+
+
 
 SCM STk_double2real(double d)
 {
@@ -1093,7 +1155,7 @@ static int isexactp(SCM z)
     case tc_integer:  return TRUE;
     default:	      error_bad_number(z);
   }
-  return STk_false; /* never reached */
+  return FALSE; /* never reached */
 }
 
 
@@ -1134,7 +1196,7 @@ doc>
       if (argc == 0) error_at_least_1();				    \
       if (_max_type_(*argv) == STk_false) error_bad_number(*argv);	    \
   									    \
-      for (previous = *argv++; --argc; previous = *argv++) {		    \
+      for (previous = *argv--; --argc; previous = *argv--) {		    \
         if (_max_type_(*argv) == STk_false) error_bad_number(*argv);	    \
 	if (do_compare(previous, *argv) _operator_ 0) return STk_false;     \
       }									    \
@@ -1160,11 +1222,11 @@ COMPARE_NUM(">=", numge, STk_realp,    <)
 
 
 /* Version with only two parameters (used by runtime) STk_numeq2, STk_numgt2 ... */
-COMPARE_NUM2(numeq, STk_complexp, ==) 
-COMPARE_NUM2(numlt, STk_realp,    <)
-COMPARE_NUM2(numgt, STk_realp,    >)
-COMPARE_NUM2(numle, STk_realp,    <=)
-COMPARE_NUM2(numge, STk_realp,    >=)
+COMPARE_NUM2(numeq,   STk_complexp, ==) 
+COMPARE_NUM2(numlt,   STk_realp,    <)
+COMPARE_NUM2(numgt,   STk_realp,    >)
+COMPARE_NUM2(numle,   STk_realp,    <=)
+COMPARE_NUM2(numge,   STk_realp,    >=)
 
 
 
@@ -1196,7 +1258,7 @@ static Inline int is_oddp(SCM n)
     		     }
     default:	     error_bad_number(n);
   }
-  return STk_void;	/* never reached */
+  return FALSE;	/* never reached */
 }
 
 static int zerop(SCM n)
@@ -1209,7 +1271,7 @@ static int zerop(SCM n)
     case tc_integer:  return (INT_VAL(n) == 0);
     default:	      error_bad_number(n);
   }
-  return STk_void;	/* never reached */
+  return FALSE;	/* never reached */
 }
 
 static int positivep(SCM n)
@@ -1222,7 +1284,7 @@ static int positivep(SCM n)
     case tc_complex:  error_not_on_a_complex(n); break;
     default:	      error_bad_number(n);
   }
-  return STk_void;	/* never reached */
+  return FALSE;	/* never reached */
 }
 
 
@@ -1236,7 +1298,7 @@ static int negativep(SCM n)
     case tc_complex:  error_not_on_a_complex(n); break;
     default:	      error_bad_number(n);
   }
-  return STk_void;	/* never reached */
+  return FALSE;	/* never reached */
 }
 
 
@@ -1305,7 +1367,7 @@ DEFINE_PRIMITIVE("max", max, vsubr, (int argc, SCM *argv))
 
   exactp = isexactp(*argv);
 
-  for (res = *argv++; --argc; argv++) {
+  for (res = *argv--; --argc; argv--) {
     /* See that the argument is a correct number */
     if (STk_realp(*argv) == STk_false) error_bad_number(*argv);
 
@@ -1333,7 +1395,7 @@ DEFINE_PRIMITIVE("min", min, vsubr, (int argc, SCM *argv))
 
   exactp = isexactp(*argv);
 
-  for (res = *argv++; --argc; argv++) {
+  for (res = *argv--; --argc; argv--) {
     /* See that the argument is a correct number */
     if (STk_realp(*argv) == STk_false) error_bad_number(*argv);
 
@@ -1378,7 +1440,8 @@ static SCM add2(SCM o1, SCM o2)
 	}
       case tc_integer: 
 	{
-	  long add = (long) INT_VAL(o1) + INT_VAL(o2);
+	  long add =  (long) INT_VAL(o1) + INT_VAL(o2);
+
 	  if (LONG_FITS_INTEGER(add))
 	    o1 = MAKE_INT(add);
 	  else
@@ -1420,7 +1483,7 @@ DEFINE_PRIMITIVE("+", plus, vsubr, (int argc, SCM *argv))
   if (argc == 0) return MAKE_INT(0);
   if (argc == 1) return add2(MAKE_INT(0), *argv);
 
-  for (res = *argv++; --argc; argv++)
+  for (res = *argv--; --argc; argv--)
     res = add2(res, *argv);
 
   return res;
@@ -1499,7 +1562,7 @@ DEFINE_PRIMITIVE("*", multiplication, vsubr, (int argc, SCM *argv))
   if (argc == 0) return MAKE_INT(1);
   if (argc == 1) return mul2(MAKE_INT(1), *argv);
 
-  for (res = *argv++; --argc; argv++)
+  for (res = *argv--; --argc; argv--)
     res = mul2(res, *argv);
   
   return res;
@@ -1594,7 +1657,7 @@ DEFINE_PRIMITIVE("-", difference, vsubr, (int argc, SCM *argv))
   if (argc == 0) error_at_least_1();
   if (argc == 1) return sub2(MAKE_INT(0), *argv);
 
-  for (res = *argv++ ; --argc; argv++)
+  for (res = *argv-- ; --argc; argv--)
     res = sub2(res, *argv);
   return res;
 }    
@@ -1620,7 +1683,7 @@ static SCM div2(SCM o1, SCM o2)
       break;
     case tc_real:
       {
-	long double r2 = REAL_VAL(o2);
+	double r2 = REAL_VAL(o2);
 	
 	if (r2 == 0.0)
 	  error_divide_by_0(o1);
@@ -1662,10 +1725,10 @@ DEFINE_PRIMITIVE("/", division, vsubr, (int argc, SCM *argv))
   if (argc == 0) error_at_least_1();
   if (argc == 1) return div2(MAKE_INT(1), *argv);
 
-  for (res = *argv++; --argc; argv++)
+  for (res = *argv--; --argc; argv--)
     res = div2(res, *argv);
   return res;
-}    
+}
 
 SCM STk_div2(SCM o1, SCM o2)
 {
@@ -1804,8 +1867,8 @@ static void integer_division(SCM x, SCM y, SCM *quotient, SCM* remainder)
     *quotient  = bignum2number(q);
     *remainder = bignum2number(r);
   } else {
-    *quotient  = mpz_get_d(q);
-    *remainder = mpz_get_d(r);
+    *quotient  = double2real(mpz_get_d(q));
+    *remainder = double2real(mpz_get_d(r));
   }
   mpz_clear(q); mpz_clear(r);                          /* //FIXME: TESTER */
 }
@@ -1879,7 +1942,7 @@ DEFINE_PRIMITIVE("gcd", gcd, vsubr, (int argc, SCM *argv))
   if (argc == 0) return MAKE_INT(0);
   if (argc == 1) return gcd2(*argv, MAKE_INT(0));
 
-  for (res = *argv++; --argc; argv++)
+  for (res = *argv--; --argc; argv--)
     res = absolute(gcd2(res, *argv));
   
   return res;
@@ -1893,7 +1956,7 @@ DEFINE_PRIMITIVE("lcm", lcm, vsubr, (int argc, SCM *argv))
   if (argc == 0) return MAKE_INT(1);
   if (STk_numberp(*argv) == STk_false) error_bad_number(*argv);
 
-  for (res = *argv++; --argc; argv++) {
+  for (res = *argv--; --argc; argv--) {
     gcd = gcd2(res, *argv);
     res = mul2(res,div2(*argv, gcd));
   }
@@ -2331,13 +2394,13 @@ static SCM my_atan2(SCM y, SCM x)
      return my_##name(z); 				\
   }
 
-transcendental(exp);
-transcendental(log);
-transcendental(sin);
-transcendental(cos);
-transcendental(tan);
-transcendental(asin);
-transcendental(acos);
+transcendental(exp)
+transcendental(log)
+transcendental(sin)
+transcendental(cos)
+transcendental(tan)
+transcendental(asin)
+transcendental(acos)
 
 DEFINE_PRIMITIVE("atan", atan, subr12, (SCM y, SCM x))
 {
@@ -2729,8 +2792,115 @@ DEFINE_PRIMITIVE("string->number", string2number, subr12, (SCM str, SCM base))
 }
  
 
+/*
+ *
+ * Bit Manipulation procedures 
+ *
+ */
+
+DEFINE_PRIMITIVE("%bit-or", bit_or, subr2, (SCM n1, SCM n2))
+{
+  mpz_t n;
+  SCM z;
+
+  switch (TYPEOF(n1)) {
+    case tc_bignum:  /* nothing */ break;
+    case tc_integer: n1 = long2scheme_bignum(INT_VAL(n1)); break;
+    default:	     error_bad_number(n1);
+  }
+
+  switch (TYPEOF(n2)) {
+    case tc_bignum:  /* nothing */ break;
+    case tc_integer: n2 = long2scheme_bignum(INT_VAL(n2)); break;
+    default:	     error_bad_number(n2);
+  }
+
+  mpz_init(n);
+  mpz_ior(n, BIGNUM_VAL(n1), BIGNUM_VAL(n2));
+  z = bignum2number(n);
+  mpz_clear(n);
+
+  return z;
+}
+
+DEFINE_PRIMITIVE("%bit-and", bit_and, subr2, (SCM n1, SCM n2))
+{
+  mpz_t n;
+  SCM z;
+
+  switch (TYPEOF(n1)) {
+    case tc_bignum:  /* nothing */ break;
+    case tc_integer: n1 = long2scheme_bignum(INT_VAL(n1)); break;
+    default:	     error_bad_number(n1);
+  }
+
+  switch (TYPEOF(n2)) {
+    case tc_bignum:  /* nothing */ break;
+    case tc_integer: n2 = long2scheme_bignum(INT_VAL(n2)); break;
+    default:	     error_bad_number(n2);
+  }
+  mpz_init(n);
+  mpz_and(n, BIGNUM_VAL(n1), BIGNUM_VAL(n2));
+  z = bignum2number(n);
+  mpz_clear(n);
+
+  return z;
+}
+
+DEFINE_PRIMITIVE("%bit-xor", bit_xor, subr2, (SCM n1, SCM n2))
+{
+  mpz_t tmp1, tmp2, tmp3;
+  SCM z;
+
+  switch (TYPEOF(n1)) {
+    case tc_bignum:  /* nothing */ break;
+    case tc_integer: n1 = long2scheme_bignum(INT_VAL(n1)); break;
+    default:	     error_bad_number(n1);
+  }
+
+  switch (TYPEOF(n2)) {
+    case tc_bignum:  /* nothing */ break;
+    case tc_integer: n2 = long2scheme_bignum(INT_VAL(n2)); break;
+    default:	     error_bad_number(n2);
+  }
+
+  /* mpz_xor exists only for version 3.x of GMP => do it by hand to allow 
+   * compilation with old versions of GMP
+   */
+  mpz_init(tmp1); mpz_init(tmp2); mpz_init(tmp3);
+  mpz_and(tmp1, BIGNUM_VAL(n1), BIGNUM_VAL(n2));
+  mpz_com(tmp2, tmp1);
+  mpz_ior(tmp1, BIGNUM_VAL(n1), BIGNUM_VAL(n2));
+  mpz_and(tmp3, tmp1, tmp2);
+  z = bignum2number(tmp3);
+  mpz_clear(tmp1); mpz_clear(tmp2); mpz_clear(tmp3);
+
+  return z;
+}
 
 
+DEFINE_PRIMITIVE("bit-not", bit_not, subr1, (SCM n))
+{
+  mpz_t z;
+
+  switch (TYPEOF(n)) {
+    case tc_bignum:  /* nothing */ break;
+    case tc_integer: n = long2scheme_bignum(INT_VAL(n)); break;
+    default:	     error_bad_number(n);
+  }
+
+  mpz_init(z);
+  mpz_com(z, BIGNUM_VAL(n));
+  
+  return bignum2number(z);
+}
+
+
+/*
+ *
+ * Initialization
+ *
+ */
 int STk_init_number(void)
 {
 #if defined(__linux__) && defined(__alpha__)
@@ -2826,5 +2996,10 @@ int STk_init_number(void)
 
   ADD_PRIMITIVE(number2string);
   ADD_PRIMITIVE(string2number);
+
+  ADD_PRIMITIVE(bit_or);
+  ADD_PRIMITIVE(bit_and);
+  ADD_PRIMITIVE(bit_xor);
+  ADD_PRIMITIVE(bit_not);
   return TRUE;
 }
