@@ -2,7 +2,7 @@
  *
  *  o b j e c t . c			-- Objects support
  *
- * Copyright  1994-2000 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
+ * Copyright  1994-2002 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
  * 
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -21,7 +21,7 @@
  *
  *            Author: Erick Gallesio [eg@unice.fr]
  *    Creation date:  9-Feb-1994 15:56
- * Last file update: 30-Dec-2000 10:05 (eg)
+ * Last file update: 27-Dec-2002 22:25 (eg)
  */
 
 #include "stklos.h"
@@ -57,7 +57,7 @@ static SCM unused; 	/* for the (useless here) reference returned by lookup */
 
 static SCM make_instance(SCM classe, short size, short type);
 static SCM compute_getters_n_setters(SCM slots);
-EXTERN_PRIMITIVE("class-of", class_of, subr1, (SCM obj));
+EXTERN_PRIMITIVE("class-of", class_of, subr1, (SCM obj))
 
 
 
@@ -130,51 +130,32 @@ static void error_bad_list(SCM obj)
 
 SCM STk_make_next_method(SCM gf, int argc, SCM *argv, SCM methods)
 {
-  SCM z;
+  SCM z, *p;
+  int i;
 
-  NEWCELL(z, next_method);
+  /* Copy parameters in the next-method object (in fact objects are not always 
+   * on the stack due to tail recursive code handling). 
+   */
+  NEWCELL_WITH_LEN(z, next_method, NXT_MTHD_ALLOC_BYTES(argc));
   NXT_MTHD_GF(z)      = gf;
-  NXT_MTHD_ARGC(z)    = argc;
-  NXT_MTHD_ARGV(z)    = argv;
   NXT_MTHD_METHOD(z)  = CAR(methods);
   NXT_MTHD_METHODS(z) = CDR(methods);
+  NXT_MTHD_ARGC(z)    = argc;
+  
+  for (p=NXT_MTHD_ARGV(z), i=0; i < argc; i++)
+    p[i] = *argv--;
 
   return z;
 }
 
-
-/*
- * STk_do_next_method.
- * 
- * Returns the procedure to apply. The arguments that must be passed to 
- * this procedure are stored in l
- *
- */
-SCM STk_do_next_method(SCM next_method, int argc, SCM *argv, SCM *l)
+DEFINE_PRIMITIVE("%set-next-method!", set_next_method, subr2,(SCM closure, SCM val))
 {
-  SCM m, new_nm;
-  SCM methods = NXT_MTHD_METHODS(next_method);
-
-  if (argc == 0) {
-    /* no argument given, reuse the ones of the original call */
-    argc = NXT_MTHD_ARGC(next_method);
-    argv = NXT_MTHD_ARGV(next_method);
-  }
-
-  if (NULLP(methods)) {
-    /* call (no-next-method gf method args) */
-    *l = STk_cons(NXT_MTHD_GF(next_method), 
-			STk_cons(NXT_MTHD_METHOD(next_method),
-				 LIST1(STk_argv2list(argc, argv))));
-    return GF_VAL("no-next-method");
-  }
-
-  m      = CAR(methods);
-  new_nm = STk_make_next_method(NXT_MTHD_GF(next_method),argc,argv,methods);
-
-  *l  = STk_cons(new_nm, STk_argv2list(argc, argv));
-  return INST_SLOT(m, S_procedure);
+  ENTER_PRIMITIVE(set_next_method);
+  if (!CLOSUREP(closure)) STk_error("bad procedure ~S", closure);
+  SET_NEXT_METHOD(closure, val);
+  return STk_void;
 }
+  
 
 #ifdef FIXME
 // static PRIMITIVE next_method_exists(SCM next)
@@ -273,8 +254,9 @@ static SCM sort_applicable_methods(SCM method_list, int size, SCM *targs)
   } 
   else {
     /* Too many elements in method_list to keep everything locally */
-    vector = STk_makevect(save, size);
+    vector = STk_makevect(size, save);   //FIXME:HORROR
     v      = VECTOR_DATA(vector);
+    // CE CODE DOIT tre avant la logique est faussse!!!!
   }
 
   /* Use a simple shell sort since it is generally faster than qsort on 
@@ -328,7 +310,7 @@ SCM STk_compute_applicable_methods(SCM gf, int argc, SCM *argv, int find_method)
     types = p = buffer;
   
   for (i=0; i < argc; i++)
-    *p++ = STk_class_of(*argv++);
+    *p++ = STk_class_of(*argv--);
  
   /* Build a list of all applicable methods */
   for (l = INST_SLOT(gf, S_methods); !NULLP(l); l = CDR(l)) {
@@ -354,7 +336,7 @@ SCM STk_compute_applicable_methods(SCM gf, int argc, SCM *argv, int find_method)
 }
 
 
-DEFINE_PRIMITIVE("find-method", find_method, vsubr, (SCM argc, SCM *argv))
+DEFINE_PRIMITIVE("find-method", find_method, vsubr, (int argc, SCM *argv))
 {
   SCM gf;
  
@@ -365,7 +347,7 @@ DEFINE_PRIMITIVE("find-method", find_method, vsubr, (SCM argc, SCM *argv))
   if (!GENERICP(gf))		       error_bad_generic(gf);
   if (NULLP(INST_SLOT(gf, S_methods))) error_no_method(gf);
  
-  return STk_compute_applicable_methods(gf, --argc, ++argv, TRUE);
+  return STk_compute_applicable_methods(gf, --argc, --argv, TRUE);
 }
  
 
@@ -431,7 +413,7 @@ DEFINE_PRIMITIVE("%method-more-specific?", method_more_specificp, subr3,
  *
 \*===========================================================================*/
 
-static SCM get_slot_value(SCM classe, SCM obj, SCM slot_name)
+static Inline SCM get_slot_value(SCM classe, SCM obj, SCM slot_name)
 {
   SCM l;
 
@@ -452,7 +434,7 @@ static SCM get_slot_value(SCM classe, SCM obj, SCM slot_name)
   return CALL_GF3("slot-missing", classe, obj, slot_name);
 }
 
-static SCM set_slot_value(SCM classe, SCM obj, SCM slot_name, SCM value)
+static Inline SCM set_slot_value(SCM classe, SCM obj, SCM slot_name, SCM value)
 {
   register SCM l;
 
@@ -515,21 +497,14 @@ DEFINE_PRIMITIVE("slot-ref-using-class", slot_ref_using_class, subr3,
   return (res==STk_void) ? CALL_GF3("slot-unbound", classe, obj, slot_name): res;
 }
 
-DEFINE_PRIMITIVE("slot-set-using-class!", slot_set_using_class, vsubr,
-		 (int argc, SCM argv[]))
+DEFINE_PRIMITIVE("slot-set-using-class!", slot_set_using_class, subr4,
+		 (SCM klass, SCM object, SCM slot_name, SCM value))
 {
   ENTER_PRIMITIVE(slot_set_using_class);
 
-  if (argc != 4) STk_error("bad number of parameters (it's %d and should be 4).", 
-			   argc);
-  /* argv[0] = class
-   * argv[1] = object
-   * argv[2] = slot name
-   * argv[3] = new value
-   */
-  if (!CLASSP(argv[0]))    error_bad_class(argv[0]);
-  if (!INSTANCEP(argv[1])) error_bad_instance(argv[1]);
-  return set_slot_value(argv[0], argv[1], argv[2], argv[3]);
+  if (!CLASSP(klass))     error_bad_class(klass);
+  if (!INSTANCEP(object)) error_bad_instance(object);
+  return set_slot_value(klass, object, slot_name, value);
 }
 
 
@@ -557,6 +532,19 @@ DEFINE_PRIMITIVE("%fast-slot-set!", fast_slot_set, subr3,
   return STk_void;
 }
 
+DEFINE_PRIMITIVE("%slot-ref", undoc_slot_ref, subr2, (SCM obj, SCM slot_name))
+{
+  /* As slot-ref except that we can reference an unbound slot */
+  SCM classe;
+
+  ENTER_PRIMITIVE(slot_ref);
+  if (!INSTANCEP(obj)) error_bad_instance(obj);
+
+  classe = INST_CLASS_OF(obj);
+  TEST_CHANGE_CLASS(obj, classe);
+  return get_slot_value(classe, obj, slot_name);
+}
+
 
 
 /******************************************************************************
@@ -567,9 +555,11 @@ DEFINE_PRIMITIVE("%fast-slot-set!", fast_slot_set, subr3,
 static void Inline set_slot_value_if_unbound
 			(SCM classe, SCM obj, SCM slot_name, SCM val)
 {
-  SCM old_val = get_slot_value(classe, obj, slot_name);
+#ifdef FIXME
+   SCM old_val = get_slot_value(classe, obj, slot_name);
   
-  if (old_val == STk_void) 
+  //FIXMOI:  if (old_val == STk_void) 
+#endif
     set_slot_value(classe, obj, slot_name, val);
 }
 
@@ -618,7 +608,7 @@ DEFINE_PRIMITIVE("%initialize-object", initialize_obj, subr2,(SCM obj,SCM initar
       /* set slot to its :init-form if it exists */
       tmp = CAR(CDR(CAR(get_n_set)));
       if (tmp != STk_false)
-	set_slot_value_if_unbound(classe, obj, slot_name, STk_C_apply(tmp, 0));
+	set_slot_value(classe, obj, slot_name, STk_C_apply(tmp, 1, obj));
     }
   }
   return obj;
@@ -1010,12 +1000,12 @@ DEFINE_PRIMITIVE("class-of", class_of, subr1, (SCM obj))
   /* Not instances objects */
   if (SCONSTP(obj)) {
     /* Expression is a small constant */
-    switch (obj) {
-      case STk_nil:   	return Null;
-      case STk_false: 	
-      case STk_true: 	return Boolean;
-      case STk_eof: 	return Eof;
-      default:		return UnknownClass;
+    switch (AS_LONG(obj)) {
+      case AS_LONG(STk_nil):   	return Null;
+      case AS_LONG(STk_false): 	
+      case AS_LONG(STk_true): 	return Boolean;
+      case AS_LONG(STk_eof): 	return Eof;
+      default:			return UnknownClass;
     }
   }
 
@@ -1060,7 +1050,8 @@ DEFINE_PRIMITIVE("%allocate-instance", allocate_instance, subr1, (SCM classe))
 	 (classe == Accessor)      ? TYPE_ACCESSOR 	:
     	 (classe == Simple_method) ? TYPE_SIMPLE_METHOD:
 				     TYPE_INSTANCE;
-  return make_instance(classe, MAKE_INT(INST_SLOT(classe, S_nfields)), type);
+  return make_instance(classe, INT_VAL(INST_SLOT(classe, S_nfields)), type); 
+  // FIXME:HORROR
 }
 
 
@@ -1153,6 +1144,7 @@ int STk_init_object(void)
   ADD_PRIMITIVE(classp);
   ADD_PRIMITIVE(genericp);
   ADD_PRIMITIVE(methodp);
+  ADD_PRIMITIVE(set_next_method);
   ADD_PRIMITIVE(find_method);
   ADD_PRIMITIVE(method_more_specificp);
   ADD_PRIMITIVE(class_of);
@@ -1162,6 +1154,7 @@ int STk_init_object(void)
   ADD_PRIMITIVE(slot_set_using_class);
   ADD_PRIMITIVE(fast_slot_ref);
   ADD_PRIMITIVE(fast_slot_set);
+  ADD_PRIMITIVE(undoc_slot_ref);
 
   ADD_PRIMITIVE(slot_boundp);
   ADD_PRIMITIVE(slot_existsp);
