@@ -1,7 +1,7 @@
 /*
  *  p o r t . c			-- ports implementation
  *
- * Copyright  1993-2001 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
+ * Copyright  1993-2003 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
  * 
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -20,30 +20,20 @@
  *
  *            Author: Erick Gallesio [eg@unice.fr]
  *    Creation date: 17-Feb-1993 12:27
- * Last file update: 10-Mar-2001 19:21 (eg)
+ * Last file update: 10-Feb-2003 11:18 (eg)
  *
  */
 
 #include <ctype.h>
 #include "stklos.h"
+#include "../pcre/pcreposix.h"
 #include "vm.h"
 
 #define INITIAL_LINE_SIZE 256		/* Initial size for readline */
 
-
-#if defined(WIN32) && !defined(CYGWIN32)
-#  define strcasecmp stricmp
-
-  /* Provide substitute functions dor WIN32 */
-  FILE *popen(char *cmd, char *mode) 
-  {	
-    /* Returning NULL will yield an error */
-    return NULL;
-  }
-  void pclose(FILE *f)
-  {}
-#endif
-
+static regex_t tilde_regexp;		/* used to detect if an error message 
+					 * is formatted or not */
+static SCM CrLf;			/* used in read-line only */
 
 static void error_bad_port(SCM p)
 {
@@ -146,10 +136,10 @@ DEFINE_PRIMITIVE("%set-std-port!", set_std_port, subr2, (SCM index, SCM port))
 {
   ENTER_PRIMITIVE(set_std_port);
   
-  switch (index) {
-    case MAKE_INT(0): if (!IPORTP(port)) goto badport; STk_curr_iport = port; break;
-    case MAKE_INT(1): if (!OPORTP(port)) goto badport; STk_curr_oport = port; break;
-    case MAKE_INT(2): if (!OPORTP(port)) goto badport; STk_curr_eport = port; break;
+  switch (AS_LONG(index)) {
+    case SCM_LONG(0): if (!IPORTP(port)) goto badport; STk_curr_iport = port; break;
+    case SCM_LONG(1): if (!OPORTP(port)) goto badport; STk_curr_oport = port; break;
+    case SCM_LONG(2): if (!OPORTP(port)) goto badport; STk_curr_eport = port; break;
     default: STk_error("bad code ~S", index);
   }
   return STk_void;
@@ -222,7 +212,7 @@ DEFINE_PRIMITIVE("read-char", read_char, subr01, (SCM port))
   ENTER_PRIMITIVE(read_char);
   port = verify_port(port, PORT_READ);
   c = STk_getc(port);
-  return (c == EOF) ? STk_eof : MAKE_CHARACTER((unsigned char) c);
+  return (c == EOF) ? STk_eof : MAKE_CHARACTER(c);
 }
 
 /*
@@ -253,7 +243,7 @@ DEFINE_PRIMITIVE("peek-char", peek_char, subr01, (SCM port))
   c = STk_getc(port);
   STk_ungetc(c, port);
 
-  return (c == EOF) ? STk_eof : MAKE_CHARACTER((unsigned char) c);
+  return (c == EOF) ? STk_eof : MAKE_CHARACTER(c);
 }
 
 
@@ -441,14 +431,19 @@ static int find_file_nature(SCM f)
 SCM STk_load_source_file(SCM f)
 {
   SCM sexpr;
-  SCM eval, ref;
+  SCM eval_symb, eval, ref;
   
   /* //FIXME: eval devrait tre connu sans faire de lookup(i.e. export par la VM)*/
-  eval = STk_lookup(STk_intern("eval"), STk_current_module, &ref, TRUE); 
+  eval_symb = STk_intern("eval");
 
   for ( ; ; ) {
+    /* We need to lookup for eval symbol for each expr, since it cans
+     * change while loading the file (with R5RS macro for instance, it
+     * is the case)
+     */
     sexpr = STk_read_constant(f, FALSE);
     if (sexpr == STk_eof) break;
+    eval  = STk_lookup(eval_symb, STk_current_module, &ref, TRUE);
     STk_C_apply(eval, 1, sexpr);
   }
   STk_close_port(f);
@@ -533,48 +528,56 @@ static SCM internal_format(int argc, SCM *argv, int error)
 {
   SCM port, fmt;
   int format_in_string = 0;
-  char *p;
+  char *p, *start_fmt;
   
   if (error) {
     if (argc < 1) goto Bad_list;
     format_in_string = 1;
     port = STk_open_output_string();
-    argc -= 1; 
+    argc -= 1;
   }
   else {
-    if (argc < 2) goto Bad_list;
-    port = *argv++; 
-    argc -= 2;
-
-    if (BOOLEANP(port)){
-      if (port == STk_true) port = STk_curr_oport;
-      else {
-	format_in_string = 1;
-	port = STk_open_output_string();
-      }
+    if (STRINGP(*argv)) {
+      /* This is a SRFI-28 format */
+      format_in_string = 1;
+      port = STk_open_output_string();
+      argc -= 1;
     } else {
-      verify_port(port, PORT_WRITE);
+      if (argc < 2) goto Bad_list;
+      port = *argv--; 
+      argc -= 2;
+      
+      if (BOOLEANP(port)){
+	if (port == STk_true) port = STk_curr_oport;
+	else {
+	  format_in_string = 1;
+	  port = STk_open_output_string();
+	}
+      } else {
+	verify_port(port, PORT_WRITE);
+      }
     }
   }
 
-  fmt = *argv++;
+  fmt = *argv--;
   if (!STRINGP(fmt)) STk_error("bad format string ~S", fmt);
-  
+
   /* Parse the format string */
-  for(p=STRING_CHARS(fmt); *p; p++) {
+  start_fmt = STRING_CHARS(fmt);
+  for(p = start_fmt; *p; p++) {
     if (*p == '~') {
       switch(*(++p)) {
         case 'A':
         case 'a': if (argc-- <= 0) goto TooMuch;
-	          STk_print(*argv++, port, DSP_MODE);
+	          STk_print(*argv--, port, DSP_MODE);
 		  continue;
         case 'S':
         case 's': if (argc-- <= 0) goto TooMuch;
-                  STk_print(*argv++, port, WRT_MODE);
+                  STk_print(*argv--, port, WRT_MODE);
 	          continue;        
         case 'W':
         case 'w': if (argc-- <= 0) goto TooMuch;
-	  	  STk_print_star(*argv++, port);
+	  	  STk_print_star(*argv--, port);
 	          continue;
         case '%': STk_putc('\n', port);
                   continue;
@@ -588,22 +591,22 @@ static SCM internal_format(int argc, SCM *argv, int error)
   }
 
   /* Verify that it doesn't remain arguments on the list */
-  if (argc) STk_error("too few ``~~'' in format string for parameter list ~S", 
-		      STk_list(argc, argv));
+  if (argc) 
+    STk_error("too few ``~~'' in format string %S", start_fmt);
 
   return format_in_string ? STk_get_output_string(port) : STk_void;
 
 TooMuch:
-  STk_error("too many ``~~'' in format string for parameter list ~S", 
-	    STk_list(argc, argv));
+  STk_error("too many ``~~'' in format string %S", start_fmt);
 Bad_list:
-  STk_error("bad list of parameters ~S", STk_list(argc, argv));
+  STk_error("bad list of parameters ~S", *argv);
   return STk_void;
 }
 
 /*
 <doc ext format 
  * (format port str obj ...)
+ * (format str obj)
  *
  * Writes the |obj|s to the given |port|, according to the format
  * string |str|. |Str| is written literally, except for the following 
@@ -634,6 +637,10 @@ Bad_list:
  *    (format #f "A ~a." "test")  => "A test."
  *    (format #f "A ~s." "test")  => "A \"test\"."
  * @end lisp
+ * 
+ * The second form of |format| is compliant with SRFI-28. That is, when
+ * |port| is omitted, the output is returned as a string as if |port| was 
+ * given the value |#f|.
 doc>
  */
 DEFINE_PRIMITIVE("format", format, vsubr, (int argc, SCM *argv))
@@ -648,32 +655,72 @@ DEFINE_PRIMITIVE("format", format, vsubr, (int argc, SCM *argv))
  * (error str obj ...)
  * (error name str obj ...)
  *
- * |error| is used to signal an error to the user. This procedure builds an 
- * error message according to the specification given in |str|. The
- * specification string follows the @emph{``tilde conventions''} of 
- * |format| (@pxref{format}). The second form of |error| takes 
- * a symbol as first parameter; it is generally used for the name of the 
- * procedure which raises the error.
+ * |error| is used to signal an error to the user. The second form 
+ * of |error| takes  a symbol as first parameter; it is generally used for the 
+ * name of the procedure which raises the error.
+ *
+ * @b{Note}: The specification string may follow the @emph{``tilde conventions''} 
+ * of |format| (@pxref{format}); in this case this procedure builds an 
+ * error message according to the specification given in |str|. Otherwise, 
+ * this procedure is conform to the |error| procedure defined in SRFI-23 and 
+ * |str| is printed with the |display| procedure, whereas the |obj|s are printed 
+ * with the |write| procedure. 
+ *
+ * @noindent
+ * Hereafter, are some calls of the |error| procedure using a formatted string
  * @lisp
- * (error "bad integer ~S" 'a)              @print{} bad integer a
- * (error 'vector-ref "bad integer ~S" 'a)  @print{} vactor-ref: bad integer a
+ * (error "bad integer ~A" "a")
+ *                      @print{} bad integer a
+ * (error 'vector-ref "bad integer ~S" "a") 
+ *                      @print{} vector-ref: bad integer "a"
+ * (error 'foo "~A is not between ~A and ~A" "bar" 0 5)
+ *                      @print{} foo: bar is not between 0 and 5
+ * @end lisp
+ *
+ * @noindent
+ * and some conform to SRFI-23
+ * @lisp
+ * (error "bad integer" "a")
+ *                     @print{} bad integer "a"
+ * (error 'vector-ref "bad integer" "a")
+ *                    @print{} vector-ref: bad integer "a"
+ * (error "bar" "is not between" 0 "and" 5)
+ *                    @print{} bar "is not between" 0 "and" 5
  * @end lisp
 doc>
  */
+static SCM srfi_23_error(int argc, SCM *argv)
+{
+  SCM port = STk_open_output_string();
+  
+  STk_print(*argv--, port, DSP_MODE); /* the message (we know that it exists) */
+  for (argc--; argc; argc--) {
+    STk_putc(' ', port);
+    STk_print(*argv--, port, WRT_MODE);
+  }
+  STk_close_port(port);
+  return STk_get_output_string(port);
+}
+
 DEFINE_PRIMITIVE("error", scheme_error, vsubr, (int argc, SCM *argv))
 {
-  char *p_name = STk_primitive_name; /* use a local cause internal_format can 
-				      * broke it */
+  regmatch_t pm[10];
+  char *p_name = ""; /* use a local cause internal_format can  broke it */
+
   if (argc > 0) {
     if (SYMBOLP(*argv)) {
       p_name = SYMBOL_PNAME(*argv);
       argc -= 1;
-      argv += 1;
-    } else {
-      p_name = "";
+      argv -= 1;
     }
     if (argc > 0) {
-      SCM msg = internal_format(argc, argv, TRUE);
+      SCM msg;
+      
+      /* See if we have a formatted message or a plain SRFI-23 call */
+      if (STRINGP(*argv) && regexec(&tilde_regexp, STRING_CHARS(*argv), 10, pm, 0))
+	msg = srfi_23_error(argc, argv);
+      else 
+	msg = internal_format(argc, argv, TRUE);
       STk_signal_error(p_name, msg);
     }
   }
@@ -750,10 +797,10 @@ DEFINE_PRIMITIVE("port-closed?", port_closed, subr1, (SCM port))
  * Reads the next line available from the input port |port|. This function
  * returns 2 values: the first one is is the string which contains the line
  * read, and the second one is the end of line delimiter. The end of line
- * demimiter can be an end of file object, a character or a list of
- * characters.  If no more characters are available on |port|, an end of
- * file object is returned.  |Port| may be omitted, in which case it
- * defaults to the value returned by |current-input-port|.
+ * delimiter can be an end of file object, a character or a string in case 
+ * of a multiple character delimiter. If no more characters are available 
+ * on |port|, an end of file object is returned.  |Port| may be omitted, 
+ * in which case it defaults to the value returned by |current-input-port|.
  *
  * @strong{Note}: As said in @xref{Multiple values}, if |read-line| is not
  * used in  the context of |call-with-values|, the second value return by 
@@ -762,14 +809,15 @@ doc>
 */
 DEFINE_PRIMITIVE("read-line", read_line, subr01, (SCM port))
 {
-  int c;
+  int prev, c;
   char buffer[INITIAL_LINE_SIZE], *buff;
   size_t i, size = INITIAL_LINE_SIZE;
-  SCM res;
+  SCM res, delim;
 
   ENTER_PRIMITIVE(read_line);
   port = verify_port(port, PORT_READ);
   buff = buffer;
+  prev = ' ';
 
   for (i = 0; ; i++) {
     if (i == size) {
@@ -784,15 +832,20 @@ DEFINE_PRIMITIVE("read-line", read_line, subr01, (SCM port))
 	buff = STk_must_realloc(buff, size);
     }
     switch (c = STk_getc(port)) {
-      case EOF:  if (i == 0) return STk_2_values(STk_eof, STk_eof);
-		 /* NO BREAK */
-
-      case '\n': res = STk_chars2string(buff, i);
+      case EOF:  res = (i == 0) ? STk_eof : STk_chars2string(buff, i);
 		 if (buff != buffer) STk_free(buff);
-		 return STk_2_values(res, 
-				     (c == EOF) ? STk_eof: MAKE_CHARACTER('\n'));
+		 return STk_2_values(res, STk_eof);
 
-      default:   buff[i] = c;
+      case '\n': if (prev == '\r') 
+		   { i -= 1; delim = CrLf; }
+      		 else 
+		   delim = MAKE_CHARACTER('\n');
+	
+		 res = STk_chars2string(buff, i);
+		 if (buff != buffer) STk_free(buff);
+		 return STk_2_values(res, delim);
+
+      default:  buff[i] = prev = c; 
     }
   }
 }
@@ -879,6 +932,17 @@ struct extended_type_descr xtype_port = {
 
 int STk_init_port(void)
 {
+  /* Initialize tilde_regexp which is used for detecting formatted error
+   * messages 
+   */
+  regcomp(&tilde_regexp, "~(a|A|s|S|w|W|~|%)", 0);
+
+  /* Define a constant for lines terminated by CR/LF to avoid multiple 
+   * allocations. Make it constant to avoid the user break it 
+   */
+  CrLf		       = STk_Cstring2string("\r\n");
+  BOXED_INFO(CrLf)    |= STRING_CONST;
+
   /* Define the port file */
   DEFINE_XTYPE(port, &xtype_port);
 
