@@ -1,7 +1,7 @@
 /*
  * stklos.h	-- stklos.h
  * 
- * Copyright  1999-2001 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
+ * Copyright  1999-2003 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
  * 
  * 
  * This program is free software; you can redistribute it and/or modify
@@ -21,7 +21,7 @@
  * 
  *           Author: Erick Gallesio [eg@unice.fr]
  *    Creation date: 28-Dec-1999 22:58 (eg)
- * Last file update:  7-Mar-2001 16:33 (eg)
+ * Last file update: 10-Feb-2003 23:51 (eg)
  */
 
 #ifndef STKLOS_H
@@ -46,6 +46,7 @@ extern "C"
 
 #include "stklosconf.h"
 #include "extraconf.h"
+
 /*===========================================================================*\
  * 
  * 		Declaration of some constants (mainly maxima) 
@@ -69,15 +70,12 @@ extern "C"
 #define MAX_TOKEN_SIZE 		1024		/* max size of a token */
 #define MAX_CHAR_CODE		255		/* Max code for a char */
 
-/*===========================================================================*\
- * 
- * 	GCC specific stuff
- * 
-\*===========================================================================*/
 
 #define CPP_CONCAT(x, y) 	x##y
 #define Inline inline 
 
+#define AS_LONG(x)		((unsigned long) (x))
+#define AS_SCM(x)		((SCM) ((unsigned long) (x)))
 
 
 /*===========================================================================*\
@@ -140,7 +138,7 @@ GC_API char *GC_stackbottom;
 
 #define MAX_CELL_TYPES		256
 
-typedef unsigned long SCM;
+typedef void* SCM;
 
 typedef enum {
   tc_not_boxed=-1,						
@@ -148,9 +146,9 @@ typedef enum {
   tc_complex, tc_symbol, tc_keyword, tc_string, tc_module, 	/* 5 */
   tc_instance, tc_closure, tc_subr0, tc_subr1, tc_subr2, 	/* 10 */
   tc_subr3, tc_subr4, tc_subr5, tc_subr01, tc_subr12,  		/* 15 */
-  tc_subr23, tc_vsubr, tc_apply, tc_vector, tc_hash_table,	/* 20 */
-  tc_port, tc_frame, tc_next_method, tc_promise, tc_regexp,	/* 25 */
-  tc_process, tc_continuation, tc_values,			/* 30 */
+  tc_subr23, tc_vsubr, tc_apply, tc_vector, tc_uvector,		/* 20 */
+  tc_hash_table, tc_port, tc_frame, tc_next_method, tc_promise, /* 25 */
+  tc_regexp, tc_process, tc_continuation, tc_values,		/* 30 */
   tc_last_standard /* must be last as indicated by its name */
 } type_cell;
 
@@ -165,8 +163,8 @@ typedef enum {
    *     .........11		small constant (#t #f '() ... see below for details)
    */
 
-#define MAKE_SCONST(n)   (n << 2 | 3)
-#define SCONSTP(n)       ((n & 0x3) == 3)
+#define MAKE_SCONST(n)   (AS_SCM(n << 2 | 3))
+#define SCONSTP(n)       ((AS_LONG(n) & 0x3) == 3)
 
   /* 
    * Header which must always be put in front of the various boxed types
@@ -185,7 +183,7 @@ typedef struct {
 
 #define BOXED_TYPE(x)	 	(((stk_header *) x)->type)
 #define BOXED_INFO(x)	 	(((stk_header *) x)->cell_info)
-#define BOXED_OBJP(x)		(!(x & 3))
+#define BOXED_OBJP(x)		(!(AS_LONG(x) & 3))
 #define BOXED_TYPE_EQ(x, y)	(BOXED_OBJP(x) && BOXED_TYPE(x) == y)
 
 #define STYPE(x)		(BOXED_OBJP(x)? BOXED_TYPE(x): tc_not_boxed)
@@ -252,6 +250,7 @@ struct primitive_obj {
 #define ADD_PRIMITIVE(_name)   STk_add_primitive(CPP_CONCAT(&o_, _name))
 #define THE_PRIMITIVE(_name)   ((SCM) CPP_CONCAT(&o_, _name))
 
+#define ENTER_PROCEDURE(_name) {STk_primitive_name = _name; }
 
 /*
   ------------------------------------------------------------------------------
@@ -287,9 +286,9 @@ int STk_init_boolean(void);
    * machine)
    */
 
-#define MAKE_CHARACTER(n) (n <<3 | 0x6)
-#define CHARACTER_VAL(n)  ((unsigned char) (n >> 3))
-#define CHARACTERP(n)	  ((n & 7) == 6)
+#define MAKE_CHARACTER(n) (AS_SCM((n) << 3 | 0x6))
+#define CHARACTER_VAL(n)  ((unsigned char) (AS_LONG(n) >> 3))
+#define CHARACTERP(n)	  ((AS_LONG(n) & 7) == 6)
 
 char *STk_char2string(unsigned char c);
 unsigned char STk_string2char(char *s);
@@ -303,7 +302,6 @@ int STk_init_char(void);
   ------------------------------------------------------------------------------
 */
 
-
 #define MODULE_ENTRY_START(_name)  				\
     void STk_module_main(void)	{				\
       static int __already_loaded = 0;				\
@@ -329,7 +327,7 @@ SCM STk_load_object_file(SCM f, char *fname);
 
 extern char *STk_primitive_name;
 
-  void STk_signal_error(char *primitive_name, SCM str);
+void STk_signal_error(char *primitive_name, SCM str);
 void STk_error(char *format, ...);
 void STk_warning(char *format, ...);
 void STk_panic(char *format, ...);
@@ -423,8 +421,10 @@ struct keyword_obj {
 #define KEYWORD_PNAME(p) 	(((struct keyword_obj *) (p))->pname)
 #define KEYWORDP(p)		(BOXED_TYPE_EQ((p),tc_keyword))
 
+#define KEYWORD_NEEDS_BARS 	(1 << 0)	/* Info flag */
+
 EXTERN_PRIMITIVE("key-set!", key_set, subr3, (SCM l, SCM key, SCM val))
-EXTERN_PRIMITIVE("key-get", key_get, subr23, (SCM l, SCM key, SCM dflt));
+EXTERN_PRIMITIVE("key-get", key_get, subr23, (SCM l, SCM key, SCM dflt))
 
 SCM STk_makekey(char *tok);
 int STk_init_keyword(void);
@@ -474,8 +474,10 @@ struct cons_obj {
 
 
 int STk_int_length(SCM l); 		/* len of a list -1 if badly formed */
-  SCM STk_argv2list(int argc, SCM *argv);
+SCM STk_argv2list(int argc, SCM *argv);
 SCM STk_append2(SCM l1, SCM l2);
+SCM STk_dappend2(SCM l1, SCM l2);	/* destructive append */
+SCM STk_dremq(SCM obj, SCM list);	/* destructive remove with eq? */
 
 EXTERN_PRIMITIVE("cons", cons, subr2, (SCM x, SCM y))
 EXTERN_PRIMITIVE("car", car, subr1, (SCM x))
@@ -525,8 +527,9 @@ int STk_init_misc(void);
    */
 
 #define INT_VAL(p)	(((long) p) >> 2)
-#define INTP(p)		(((p) & 3) == 1)
-#define MAKE_INT(n)	(((n) << 2) | 1)
+#define INTP(p)		((((unsigned long) p) & 3) == 1)
+#define SCM_LONG(n)	(((n) << 2) | 1)
+#define MAKE_INT(n)	(AS_SCM(SCM_LONG(n)))
 #define INT_MIN_VAL	((LONG_MIN & ~3) >> 2)
 #define INT_MAX_VAL	((LONG_MAX & ~3) >> 2)
 
@@ -538,7 +541,7 @@ unsigned long STk_uinteger_value(SCM x); /* Returns ULONG_MAX if not an ulong */
    **** Real 
    ****/
 
-#define REAL_FORMAT_SIZE         15
+#define REAL_FORMAT_SIZE         17
 
 struct real_obj {
   stk_header header;
@@ -591,13 +594,14 @@ struct complex_obj {
   /**** 
    **** Conversions 
    ****/
-SCM    STk_Cstr2number(char *str, long base);
-char  *STk_bignum2Cstring(SCM n, int base);
-SCM    STk_long2integer(long n);
-SCM    STk_ulong2integer(unsigned long n);
-SCM    STk_double2real(double d);
-int    STk_init_number(void);
-double STk_number2double(SCM n);
+SCM    		STk_Cstr2number(char *str, long base);
+char  	       *STk_bignum2Cstring(SCM n, int base);
+SCM    		STk_long2integer(long n);
+SCM    		STk_ulong2integer(unsigned long n);
+SCM    		STk_double2real(double d);
+double 		STk_number2double(SCM n);
+long   	      	STk_integer2int32(SCM n, int *overflow);
+unsigned long 	STk_integer2uint32(SCM n, int *overflow);
 
   /****
    **** Arithmetic
@@ -608,12 +612,16 @@ SCM STk_mul2(SCM o1, SCM o2);
 SCM STk_div2(SCM o1, SCM o2);
 
 long STk_numeq2(SCM o1, SCM o2);
+long STk_numdiff2(SCM o1, SCM o2);
 long STk_numlt2(SCM o1, SCM o2);
 long STk_numgt2(SCM o1, SCM o2);
 long STk_numle2(SCM o1, SCM o2);
 long STk_numge2(SCM o1, SCM o2);
 
 
+int    STk_init_number(void);
+
+
 #define NUMBERP(x)	(INTP(x) || BIGNUMP(x) || REALP(x) || RATIONALP(x) || \
 		         COMPLEXP(x))
 #define EXACTP(x)	(INTP(x) || BIGNUMP(x) || EXACT_RATIONALP(x) || \
@@ -638,6 +646,9 @@ int STk_init_object(void);
   ------------------------------------------------------------------------------
 */
 char *STk_expand_file_name(char *s);
+SCM STk_do_glob(int argc, SCM *argv);
+SCM STk_resolve_link(char *path, int count);
+
 
 /*
   ------------------------------------------------------------------------------
@@ -667,6 +678,7 @@ struct port_obj {
   int  pos;			/* position from the start of file (4 read only) */
   /* virtual functions (in the object 'cause the # of ports should be low ) */
   void (*print_it)  (SCM obj, SCM port);  /* used to display or print object */
+  void (*release_it)(SCM obj);
   int  (*creadyp)   (void *stream); 
   int  (*cgetc)     (void *stream);
   int  (*ceofp)     (void *stream);
@@ -697,6 +709,7 @@ struct port_obj {
 #define PORT_POS(x)	(((struct port_obj *) (x))->pos)
 
 #define PORT_PRINT(x)   (((struct port_obj *) (x))->print_it)
+#define PORT_RELEASE(x) (((struct port_obj *) (x))->release_it)
 #define PORT_READY(x)	(((struct port_obj *) (x))->creadyp)
 #define PORT_GETC(x)	(((struct port_obj *) (x))->cgetc)
 #define PORT_EOFP(x)	(((struct port_obj *) (x))->ceofp)
@@ -744,11 +757,13 @@ int STk_write_buffer(SCM port, void *buff, int count);
  ****/
 SCM STk_rewind_file_port(SCM port);
 SCM STk_open_file(char *filename, char *mode);
-SCM STk_set_port_idle(SCM port, SCM idle_func);
+SCM STk_add_port_idle(SCM port, SCM idle_func);
 SCM STk_fd2scheme_port(int fd, char *mode, char *identification);
+void STk_set_line_buffered_mode(SCM port);
 int STk_init_fport(void);
 
 
+
 /****
  ****		sport.h primitives
  ****/
@@ -824,7 +839,7 @@ struct closure_obj {
   stk_header header;
   short arity;
   unsigned short code_size;
-  //  SCM formals;
+  /*  SCM formals; */
   SCM env;
   SCM plist;
   SCM name;
@@ -832,9 +847,10 @@ struct closure_obj {
   STk_instr *bcode;
 };
 
-
-//#define CLOSURE_FORMALS(p)	(((struct closure_obj *) (p))->formals)
-//#define CLOSURE_CODE(p)   	(((struct closure_obj *) (p))->code)
+/* FIXME:
+#define CLOSURE_FORMALS(p)	(((struct closure_obj *) (p))->formals) 
+#define CLOSURE_CODE(p)   	(((struct closure_obj *) (p))->code)
+*/
 
 #define CLOSURE_ARITY(p)	(((struct closure_obj *) (p))->arity)
 #define CLOSURE_SIZE(p)		(((struct closure_obj *) (p))->code_size)
@@ -845,7 +861,7 @@ struct closure_obj {
 #define CLOSURE_BCODE(p)	(((struct closure_obj *) (p))->bcode)
 #define CLOSUREP(p)		(BOXED_TYPE_EQ((p), tc_closure))
 
-EXTERN_PRIMITIVE("procedure?", procedurep, subr1, (SCM obj));
+EXTERN_PRIMITIVE("procedure?", procedurep, subr1, (SCM obj))
 
 SCM STk_make_closure(STk_instr *code, int size, int arity, SCM *cst, SCM env);
 int STk_init_proc(void);
@@ -858,10 +874,11 @@ int STk_init_proc(void);
   ----
   ------------------------------------------------------------------------------
 */
-SCM STk_read(SCM port, int case_significant);
-SCM STk_read_constant(SCM port, int case_significant);
-int STk_init_reader(void);
-SCM STk_sym_quote;		/* The interned value of symbol "quote" */
+SCM   STk_read(SCM port, int case_significant);
+SCM   STk_read_constant(SCM port, int case_significant);
+char *STk_quote2str(SCM symb);
+int   STk_init_reader(void);
+extern SCM STk_sym_quote;		/* The interned value of symbol "quote" */
 
 
 /*
@@ -874,6 +891,15 @@ SCM STk_sym_quote;		/* The interned value of symbol "quote" */
 int STk_init_regexp(void);
 
 
+/*
+  ------------------------------------------------------------------------------
+  ----
+  ---- 				S O C K E T . C
+  ----
+  ------------------------------------------------------------------------------
+*/
+int STk_init_socket(void);
+
 /*
   ------------------------------------------------------------------------------
   ----
@@ -913,8 +939,8 @@ SCM STk_chars2string(char *str, size_t len); /* Original can have null character
 
 
 EXTERN_PRIMITIVE("string=?", streq, subr2, (SCM s1, SCM s2))
-EXTERN_PRIMITIVE("string-ref", string_ref, subr2, (SCM str, SCM index));
-EXTERN_PRIMITIVE("string-set!", string_set, subr3, (SCM str, SCM index, SCM value));
+EXTERN_PRIMITIVE("string-ref", string_ref, subr2, (SCM str, SCM index))
+EXTERN_PRIMITIVE("string-set!", string_set, subr3, (SCM str, SCM index, SCM value))
 int STk_init_string(void);
 
 /*
@@ -934,6 +960,9 @@ struct symbol_obj {
 
 #define SYMBOL_NEEDS_BARS 	(1 << 0)	/* Info flag */
 
+EXTERN_PRIMITIVE("string->symbol", string2symbol, subr1, (SCM string))
+
+int STk_need_bars(register char *s);
 SCM STk_intern(char *name);
 SCM STk_make_uninterned_symbol(char *name);
 int STk_init_symbol(void);
@@ -948,10 +977,20 @@ int STk_init_symbol(void);
 */
 
 int STk_dirp(const char *path);
-SCM STk_int_expand_file_name(char *s);
-
 int STk_init_system();
 
+/*
+  ------------------------------------------------------------------------------
+  ----
+  ---- 				 U V E C T O R . C
+  ----
+  ------------------------------------------------------------------------------
+*/
+extern int STk_uvectors_allowed;
+
+int STk_uniform_vector_tag(char *s);
+SCM STk_list2uvector(int type, SCM l);
+int STk_init_uniform_vector(void);
 /*
   ------------------------------------------------------------------------------
   ----
@@ -972,15 +1011,14 @@ struct vector_obj {
 
 #define VECTOR_CONST	 (1 << 0)
 
-EXTERN_PRIMITIVE("vector-ref", vector_ref, subr2, (SCM v, SCM index));
-EXTERN_PRIMITIVE("vector-set!", vector_set, subr3, (SCM v, SCM index, SCM value));
-EXTERN_PRIMITIVE("vector->list", vector2list, subr1, (SCM v));
+EXTERN_PRIMITIVE("vector-ref", vector_ref, subr2, (SCM v, SCM index))
+EXTERN_PRIMITIVE("vector-set!", vector_set, subr3, (SCM v, SCM index, SCM value))
+EXTERN_PRIMITIVE("vector->list", vector2list, subr1, (SCM v))
 
 SCM STk_makevect(int len, SCM init);
 SCM STk_list2vector(SCM l);
 int STk_init_vector(void);
 
-
 /*
   ------------------------------------------------------------------------------
   ----
