@@ -2,7 +2,7 @@
  *
  * s t r . c				-- Strings management
  *
- * Copyright  1993-2000 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
+ * Copyright  1993-2003 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
  * 
  *
  * This program is free software; you can redistribute it and/or modify
@@ -22,7 +22,7 @@
  *
  *           Author: Erick Gallesio [eg@unice.fr]
  *    Creation date: ??????
- * Last file update: 30-Dec-2000 10:29 (eg)
+ * Last file update: 13-Feb-2003 11:26 (eg)
  */
 
 #include <ctype.h>
@@ -87,6 +87,48 @@ static int stringcompi(SCM s1, SCM s2)
   return l1 ? +1 : (l2 ? -1 : 0);
 }
 
+static SCM control_index(int argc, SCM *argv, int *pstart, int *pend)
+{
+  SCM s = NULL;
+  int len, start=0, end=-1;
+
+  /* Controling number of arguments */
+  switch (argc) {
+  case 1: s = argv[0]; break;
+  case 2: s = argv[0]; start = STk_integer_value(argv[-1]); break;
+  case 3: s = argv[0]; start = STk_integer_value(argv[-1]); 
+	  end = STk_integer_value(argv[-2]); break;
+  default: STk_error("incorrect number of argument (%d)", argc);
+  }
+
+  /* Controlling s */
+  if (!STRINGP(s)) error_bad_string(s);
+  len = STRING_SIZE(s);
+
+  /* Controling start index */
+  if (start == LONG_MIN || start < 0 || start > len)
+    /* argc cannot be 1 (start would be 0) */
+    STk_error("bad starting index ~S", argv[(argc==2) ? 0 : -1]);
+	      
+  /* controling end index */
+  if (end == -1)
+    end = len;
+  else
+    if (end == LONG_MIN  || end < 0 || end > len)
+      /* We have an end index ==> argc = 3 */
+      STk_error("bad ending index ~S", argv[0]);
+
+  if (start > end) 
+    STk_error("low index is greater than high index");
+
+  /* everything is correct, return values */
+  *pstart = start;
+  *pend   = end;
+  return s;
+}
+
+
+
 SCM STk_makestring(int len, unsigned char *init)
 {
   register SCM z;
@@ -108,6 +150,7 @@ SCM STk_Cstring2string(char *str) /* Embed a C string in Scheme world  */
   NEWCELL_ATOMIC(z, string, sizeof(struct string_obj) + len);
   STRING_SIZE(z) = len;
   strcpy(STRING_CHARS(z), str);
+
   return z;
 }
 
@@ -185,7 +228,7 @@ DEFINE_PRIMITIVE("string", string, vsubr, (int argc, SCM* argv))
   z = STk_makestring(argc, NULL);
  
   /* copy element in newly allocated string */
-  for (s=STRING_CHARS(z); argc--; s++, argv++) {
+  for (s=STRING_CHARS(z); argc--; s++, argv--) {
     if (!CHARACTERP(*argv)) error_bad_character(*argv);
     *s = CHARACTER_VAL(*argv);
   }
@@ -227,7 +270,7 @@ DEFINE_PRIMITIVE("string-ref", string_ref, subr2, (SCM str, SCM index))
   if (k < 0 || k >= STRING_SIZE(str)) 
     STk_error("index ~S out of bound in string ~S", index, str);
   
-  return MAKE_CHARACTER((unsigned char) STRING_CHARS(str)[k]);
+  return MAKE_CHARACTER(STRING_CHARS(str)[k]);
 }
 
 
@@ -385,8 +428,9 @@ DEFINE_PRIMITIVE("string-append", string_append, vsubr, (int argc, SCM* argv))
 
   /* Compute total length of resulting string */
   for (i = 0; i < argc; i++) {
-    if (!STRINGP(argv[i])) error_bad_string(argv[i]);
-    total += STRING_SIZE(argv[i]);
+    p = argv[-i];
+    if (!STRINGP(p)) error_bad_string(p);
+    total += STRING_SIZE(p);
   }
 
   /* Allocate result */
@@ -397,7 +441,7 @@ DEFINE_PRIMITIVE("string-append", string_append, vsubr, (int argc, SCM* argv))
   for (i = 0; i < argc; i++) {
     memcpy(p, STRING_CHARS(*argv), (unsigned int) STRING_SIZE(*argv));
     p    += STRING_SIZE(*argv);
-    argv +=1;
+    argv -=1;
   }
   return z;
 }
@@ -521,7 +565,7 @@ static char *Memmem(char *s1, int l1, char *s2, int l2)
 }
 
 /*
-<doc ext string-find
+<doc ext string-find?
  * (string-find? str1 str2)
  *
  * Returns |#t| if |str1| appears somewhere in |str2|; otherwise returns |#f|.
@@ -575,8 +619,8 @@ DEFINE_PRIMITIVE("string-index", string_index, subr2, (SCM s1, SCM s2))
  * |delimiters| string. If |delimiters| is omitted, it defaults to a string 
  * containing a space, a tabulation and a newline characters. 
  * @lisp
- * (split-string "/usr/local/bin" "/") => ("usr" "local" "bin")
- * (split-string "once   upon a time") => ("once" "upon" "a" "time")
+ * (string-split "/usr/local/bin" "/") => ("usr" "local" "bin")
+ * (string-split "once   upon a time") => ("once" "upon" "a" "time")
  * @end lisp
 doc>
 */
@@ -616,61 +660,220 @@ DEFINE_PRIMITIVE("string-split", string_split, subr12, (SCM string, SCM delimite
   return STk_dreverse(result);
 }
 
-
-
 /*
-<doc EXT string-lower
- * (string-lower str)
- *
- * Returns a string in which all upper case letters of string |str| have been
- * replaced by their lower case equivalent.
+<doc EXT string-mutable?
+ * (string-mutable? obj)
+ * 
+ * Returns |#t| if |obj| is a mutable string, otherwise returns |#f|. 
+ * @lisp
+ * (string-mutable? "abc")                => #f
+ * (string-mutable? (string-copy "abc"))  => #t
+ * (string-mutable? (string #\a #\b #\c)) => #t
+ * (string-mutable? 12)                   => #f
+ * @end lisp
 doc>
- */
-DEFINE_PRIMITIVE("string-lower", string_lower, subr1, (SCM s))
+*/
+DEFINE_PRIMITIVE("string-mutable?", string_mutable, subr1, (SCM obj))
 {
-  SCM z;
-  int len;
-  unsigned char *endp, *p, *q;
-
-  ENTER_PRIMITIVE(string_lower);
-  if (!STRINGP(s)) error_bad_string(s);
-
-  len  = STRING_SIZE(s);
-  endp = STRING_CHARS(s) + len;
-  z    = STk_makestring(len, NULL);
-
-  for (p=STRING_CHARS(s), q=STRING_CHARS(z); p < endp; p++, q++) *q = tolower(*p);
-  return z;
+  return MAKE_BOOLEAN(STRINGP(obj) && !(BOXED_INFO(obj) & STRING_CONST));
 }
 
 
-
 /*
-<doc EXT string-upper
- * (string-upper str)
+<doc EXT string-downcase
+ * (string-downcase str)
+ * (string-downcase str start)
+ * (string-downcase str start end)
  *
- * Returns a string in which all lower case letters of string |str| have been
- * replaced by their upper case equivalent.
+ * Returns a string in which the upper case letters of string |str| between the 
+ * |start| and |end| indices have been replaced by their lower case equivalent.
+ * If |start| is omited, it defaults to 0. If |end| is omited, it defaults to
+ * the length of |str|. 
+ * @lisp 
+ * (string-downcase "Foo BAR")        => "foo bar"
+ * (string-downcase "Foo BAR" 4)      => "bar"
+ * (string-downcase "Foo BAR" 4 6)    => "ba"
+ * @end lisp
 doc>
  */
-DEFINE_PRIMITIVE("string-upper", string_upper, subr1, (SCM s))
+DEFINE_PRIMITIVE("string-downcase", string_downcase, vsubr, (int argc, SCM *argv))
 {
-  SCM z;
-  int len;
+  SCM s, z;
+  int start, end;
   unsigned char *endp, *p, *q;
 
-  ENTER_PRIMITIVE(string_upper);
-  if (!STRINGP(s)) error_bad_string(s);
+  ENTER_PRIMITIVE(string_downcase);
 
-  len  = STRING_SIZE(s);
-  endp = STRING_CHARS(s) + len;
-  z    = STk_makestring(len, NULL);
-
-
-  for (p=STRING_CHARS(s), q=STRING_CHARS(z); p < endp; p++, q++) *q = toupper(*p);
+  s    = control_index(argc, argv, &start, &end);
+  endp = STRING_CHARS(s) + end;
+  z    = STk_makestring(end-start, NULL);
+  
+  for (p=STRING_CHARS(s)+start, q=STRING_CHARS(z); p < endp; p++, q++) 
+    *q = tolower(*p);
   return z;
 }
 
+/*
+<doc EXT string-downcase!
+ * (string-downcase! str)
+ * (string-downcase! str start)
+ * (string-downcase! str start end)
+ * 
+ * This is the in-place side-effecting variant of |string-downcase|. 
+ * @lisp
+ * (string-downcase! (string-copy "Foo BAR") 4)    => "Foo bar"
+ * (string-downcase! (string-copy "Foo BAR") 4 6)  => "Foo baR"
+ * @end lisp
+doc>
+*/
+DEFINE_PRIMITIVE("string-downcase!", string_ddowncase, vsubr, (int argc, SCM *argv))
+{
+  SCM s;
+  int start, end;
+  unsigned char *endp, *p;
+
+  ENTER_PRIMITIVE(string_downcase);
+  s    = control_index(argc, argv, &start, &end);
+  if (BOXED_INFO(s) & STRING_CONST) error_change_const_string(s);
+
+  endp = STRING_CHARS(s) + end;
+  
+  for (p=STRING_CHARS(s)+start; p < endp; p++) *p = tolower(*p);
+  return s;
+}
+
+/*
+<doc EXT string-upcase
+ * (string-upcase str)
+ * (string-upcase str start)
+ * (string-upcase str start end)
+ *
+ * Returns a string in which the lower case letters of string |str| between the 
+ * |start| and |end| indices have been replaced by their upper case equivalent.
+ * If |start| is omited, it defaults to 0. If |end| is omited, it defaults to
+ * the length of |str|. 
+doc>
+ */
+DEFINE_PRIMITIVE("string-upcase", string_upcase, vsubr, (int argc, SCM *argv))
+{
+  SCM s, z;
+  int start, end;
+  unsigned char *endp, *p, *q;
+
+  ENTER_PRIMITIVE(string_upcase);
+
+  s    = control_index(argc, argv, &start, &end);
+  endp = STRING_CHARS(s) + end;
+  z    = STk_makestring(end-start, NULL);
+  
+  for (p=STRING_CHARS(s)+start, q=STRING_CHARS(z); p < endp; p++, q++) 
+    *q = toupper(*p);
+  return z;
+}
+
+/*
+<doc EXT string-upcase!
+ * (string-upcase! str)
+ * (string-upcase! str start)
+ * (string-upcase! str start end)
+ *
+ * This is the in-place side-effecting variant of |string-upcase|. 
+doc>
+*/
+DEFINE_PRIMITIVE("string-upcase!", string_dupcase, vsubr, (int argc, SCM *argv))
+{
+  SCM s;
+  int start, end;
+  unsigned char *endp, *p;
+
+  ENTER_PRIMITIVE(string_upcase);
+  s    = control_index(argc, argv, &start, &end);
+  if (BOXED_INFO(s) & STRING_CONST) error_change_const_string(s);
+
+  endp = STRING_CHARS(s) + end;
+  
+  for (p=STRING_CHARS(s)+start; p < endp; p++) *p = toupper(*p);
+  return s;
+}
+
+
+/*
+<doc EXT string-titlecase
+ * (string-titlecase str)
+ * (string-titlecase str start)
+ * (string-titlecase str start end)
+ *
+ * This function returns a string.  For every character |c| in the
+ * selected range of |str|, if |c| is preceded by a cased character, it
+ * is downcased; otherwise it is titlecased. If |start| is omited, it 
+ * defaults to 0. If |end| is omited, it defaults to the length of |str|. 
+ * Note that if a |start| index is specified, then the character preceding
+ * |s[start]| has no effect on the titlecase decision for character |s[start]|.
+ * @lisp
+ * (string-titlecase "--capitalize tHIS sentence.") 
+ *          =>  "--Capitalize This Sentence."
+ * (string-titlecase "see Spot run. see Nix run.") 
+ *          =>  "See Spot Run. See Nix Run."
+ * (string-titlecase "3com makes routers.") 
+ *          =>  "3Com Makes Routers."
+ * (string-titlecase "greasy fried chicken" 2) 
+ *          => "Easy Fried Chicken"
+ * @end lisp
+doc>
+*/
+DEFINE_PRIMITIVE("string-titlecase", string_titlecase, vsubr, (int argc, SCM *argv))
+{
+  SCM s, z;
+  int start, end;
+  unsigned char *endp, *p, *q;
+  char prev_is_sep = 1, curr_is_sep;
+
+  ENTER_PRIMITIVE(string_upcase);
+  s    = control_index(argc, argv, &start, &end);
+  endp = STRING_CHARS(s) + end;
+  z    = STk_makestring(end-start, NULL);
+
+  for (p=STRING_CHARS(s)+start, q=STRING_CHARS(z); p < endp; p++, q++) {
+    curr_is_sep = !(isalpha(*p));
+    if (curr_is_sep)
+      *q = *p;
+    else 
+      *q = (prev_is_sep) ? toupper(*p) : tolower(*p);
+    prev_is_sep = curr_is_sep;
+  }
+  return z;
+}
+
+/*
+<doc EXT string-titlecase!
+ * (string-titlecase! str)
+ * (string-titlecase! str start)
+ * (string-titlecase! str start end)
+ *
+ * This is the in-place side-effecting variant of |string-titlecase|. 
+doc>
+*/
+DEFINE_PRIMITIVE("string-titlecase!", string_dtitlecase,vsubr,(int argc, SCM *argv))
+{
+  SCM s;
+  int start, end;
+  unsigned char *endp, *p;
+  char prev_is_sep = 1, curr_is_sep;
+
+  ENTER_PRIMITIVE(string_upcase);
+  s    = control_index(argc, argv, &start, &end);
+  endp = STRING_CHARS(s) + end;
+
+  if (BOXED_INFO(s) & STRING_CONST) error_change_const_string(s);
+
+  for (p=STRING_CHARS(s)+start; p < endp; p++) {
+    curr_is_sep = !(isalpha(*p));
+    if (!curr_is_sep)
+      *p = (prev_is_sep) ? toupper(*p) : tolower(*p);
+    prev_is_sep = curr_is_sep;
+  }
+  return s;
+}
 
 int STk_init_string(void)
 {
@@ -700,7 +903,12 @@ int STk_init_string(void)
   ADD_PRIMITIVE(string_find);
   ADD_PRIMITIVE(string_index);
   ADD_PRIMITIVE(string_split);
-  ADD_PRIMITIVE(string_lower);
-  ADD_PRIMITIVE(string_upper);
+  ADD_PRIMITIVE(string_mutable);
+  ADD_PRIMITIVE(string_downcase);
+  ADD_PRIMITIVE(string_ddowncase);
+  ADD_PRIMITIVE(string_upcase);
+  ADD_PRIMITIVE(string_dupcase);
+  ADD_PRIMITIVE(string_titlecase);
+  ADD_PRIMITIVE(string_dtitlecase);
   return TRUE;
 }
