@@ -2,7 +2,7 @@
  *
  * s y s t e m . c				-- System relative primitives
  *
- * Copyright  1994-2001 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
+ * Copyright  1994-2003 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
  * 
  *
  * Permission to use, copy, modify, distribute,and license this
@@ -16,50 +16,21 @@
  *
  *           Author: Erick Gallesio [eg@kaolin.unice.fr]
  *    Creation date: 29-Mar-1994 10:57
- * Last file update: 24-Jan-2001 09:42 (eg)
+ * Last file update:  5-Jun-2003 16:28 (eg)
  */
 
-#ifndef WIN32
-#  include <unistd.h>
-#  include <pwd.h>
-#else
-#  include <windows.h>
-#  include <io.h>
-#  define F_OK   00
-#  define X_OK   01
-#  define W_OK   02
-#  define R_OK   04
-#endif
-
+#include <unistd.h>
+#include <pwd.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-
-#ifdef WIN32
-   /* One of above files includes <stdarg.h> with BC++  (and stdarg and 
-    * vararg are not compatible 
-    */
-#  undef __STDARG_H 
-
-#  ifdef _MSC_VER
-#  include <direct.h>
-#  include <process.h>
-#  include <sys/stat.h>
-#  define S_ISDIR(mode) ((mode & _S_IFMT) == _S_IFDIR)
-#  define S_ISREG(mode) ((mode & _S_IFMT) == _S_IFREG)
-#  else
-#     ifdef BCC
-        /* Borland defines the opendir/readdir/closedir functions. Use them. */
-#       include <dirent.h>
-#     endif
-#  endif
-#else
-#  include <dirent.h>
-#endif
-
+#include <fcntl.h>
+#include <dirent.h>
 #include <time.h> 
 #include "stklos.h"
 
-
+#ifndef MAXBUFF
+#  define MAXBUFF 1024
+#endif
 
 /******************************************************************************
  *
@@ -76,6 +47,16 @@ static void error_bad_string(SCM path)
   STk_error("~S is a bad string", path);
 }
 
+static void error_bad_int_or_out_of_bounds(SCM val)
+{
+  STk_error("bad integer ~S (or out of range)", val);
+}
+
+static void error_cannot_copy(SCM f1, SCM f2)
+{
+  STk_error("cannot copy file ~S to ~S", f1, f2);
+}
+
 
 static SCM my_access(SCM path, int mode)
 {
@@ -84,7 +65,6 @@ static SCM my_access(SCM path, int mode)
 }
 
 
-
 int STk_dirp(const char *path)
 {
   struct stat buf;
@@ -95,317 +75,6 @@ int STk_dirp(const char *path)
 }
 
 
-#ifdef FIXME
-//EG: #define MAXLINK 50	/* Number max of link before declaring we have a loop */
-//EG: 
-//EG: SCM STk_resolve_link(char *path, int count)
-//EG: {
-//EG: #ifdef WIN32
-//EG:   return STk_internal_expand_file_name(path);
-//EG: #else
-//EG:   char link[MAX_PATH_LENGTH], dst[MAX_PATH_LENGTH], *s, *d=dst;
-//EG:   int n;
-//EG:   SCM p;
-//EG:   
-//EG:   p  = STk_internal_expand_file_name(path);
-//EG:   
-//EG:   for (s=CHARS(p)+1, *d++='/' ;       ; s++, d++) {
-//EG:     switch (*s) {
-//EG:       case '\0':
-//EG:       case '/' : *d = '\0';
-//EG: 	if ((n=readlink(dst, link, MAX_PATH_LENGTH-1)) > 0) {
-//EG: 	  link[n] = '\0';
-//EG: 	  if (link[0] == '/') 
-//EG: 	    /* link is absolute */
-//EG: 	    d = dst;
-//EG: 	  else {
-//EG: 	    /* relative link. Delete last item */
-//EG: 	    while (*--d != '/') {
-//EG: 	    }
-//EG: 	    d += 1;
-//EG: 	  }
-//EG: 		   
-//EG: 	  /* d points the place where the link must be placed */
-//EG: 	  if (d - dst + strlen(link) + strlen(s) < MAX_PATH_LENGTH - 1) {
-//EG: 	    /* we have enough room */
-//EG: 	    sprintf(d, "%s%s", link, s); 
-//EG: 	    /* Recurse. Be careful for loops (a->b and b->a) */
-//EG: 	    if (count < MAXLINK) 
-//EG: 	      return STk_resolve_link(dst, count+1);
-//EG: 	  }
-//EG: 	  return STk_false;
-//EG: 	}
-//EG: 	else {
-//EG: 	  if (errno != EINVAL) 
-//EG: 	    /* EINVAL = file is not a symlink (i.e. it's a true error) */
-//EG: 	    return STk_false;
-//EG: 	  else
-//EG: 	    if (*s) *d = '/'; 
-//EG: 	    else return STk_makestring(dst);		       
-//EG: 	}
-//EG:       default:   *d = *s;
-//EG:     }
-//EG:   }
-//EG: #endif
-//EG: }
-//EG: 
-//EG:    
-//EG: /*
-//EG:  *----------------------------------------------------------------------
-//EG:  *
-//EG:  * fileglob --
-//EG:  *      *****							        ******
-//EG:  * 	***** This function is an adaptation of the Tcl function DoGlob ******
-//EG:  *      ***** Adaptated to use true lists rather than string as in Tcl  ******
-//EG:  *      *****							        ******
-//EG:  *	
-//EG:  *
-//EG:  *      This recursive procedure forms the heart of the globbing
-//EG:  *      code.  It performs a depth-first traversal of the tree
-//EG:  *      given by the path name to be globbed.
-//EG:  *
-//EG:  * Results:
-//EG:  *      The return value is a standard Tcl result indicating whether
-//EG:  *      an error occurred in globbing.  After a normal return the
-//EG:  *      result in interp will be set to hold all of the file names
-//EG:  *      given by the dir and rem arguments.  After an error the
-//EG:  *      result in interp will hold an error message.
-//EG:  *
-//EG:  * Side effects:
-//EG:  *      None.
-//EG:  *
-//EG:  *----------------------------------------------------------------------
-//EG:  */
-//EG: 
-//EG: static SCM fileglob(char *dir, char *rem, SCM result)
-//EG: /* dir: Name of a directory at which to start glob expansion.  This name
-//EG:  * is fixed: it doesn't contain any globbing chars. 
-//EG:  * rem: Path to glob-expand.
-//EG:  */
-//EG: {
-//EG:   /*
-//EG:    * When this procedure is entered, the name to be globbed may
-//EG:    * already have been partly expanded by ancestor invocations of
-//EG:    * fileglob.  The part that's already been expanded is in "dir"
-//EG:    * (this may initially be empty), and the part still to expand
-//EG:    * is in "rem".  This procedure expands "rem" one level, making
-//EG:    * recursive calls to itself if there's still more stuff left
-//EG:    * in the remainder.
-//EG:    */
-//EG:   
-//EG:   Tcl_DString newName;                /* Holds new name consisting of
-//EG: 				       * dir plus the first part of rem. */
-//EG:   register char *p;
-//EG:   register char c;
-//EG:   char *openBrace, *closeBrace, *name, *dirName;
-//EG:   int gotSpecial, baseLength;
-//EG:   struct stat statBuf;
-//EG: 
-//EG:   /*
-//EG:    * Make sure that the directory part of the name really is a
-//EG:    * directory.  If the directory name is "", use the name "."
-//EG:    * instead, because some UNIX systems don't treat "" like "."
-//EG:    * automatically. Keep the "" for use in generating file names,
-//EG:    * otherwise "glob foo.c" would return "./foo.c".
-//EG:    */
-//EG:   
-//EG:   dirName = (*dir == '\0') ? ".": dir;
-//EG:   if ((stat(dirName, &statBuf) != 0) || !S_ISDIR(statBuf.st_mode))
-//EG:     return result;
-//EG: 
-//EG:   Tcl_DStringInit(&newName);
-//EG: 
-//EG:   /*
-//EG:    * First, find the end of the next element in rem, checking
-//EG:    * along the way for special globbing characters.
-//EG:    */
-//EG:   
-//EG:   gotSpecial = 0;
-//EG:   openBrace = closeBrace = NULL;
-//EG:   for (p = rem; ; p++) {
-//EG:     c = *p;
-//EG:     if ((c == '\0') || ((openBrace == NULL) && (c == '/'))) break;
-//EG:     if ((c == '{')  && (openBrace == NULL)) openBrace = p;
-//EG:     if ((c == '}')  && (openBrace != NULL) && (closeBrace == NULL)) closeBrace = p;
-//EG:     if ((c == '*') || (c == '[') || (c == '\\') || (c == '?')) gotSpecial = 1;
-//EG:   }
-//EG: 
-//EG:   /*
-//EG:    * If there is an open brace in the argument, then make a recursive
-//EG:    * call for each element between the braces.  In this case, the
-//EG:    * recursive call to fileglob uses the same "dir" that we got.
-//EG:    * If there are several brace-pairs in a single name, we just handle
-//EG:    * one here, and the others will be handled in recursive calls.
-//EG:    */
-//EG: 
-//EG:   if (openBrace != NULL) {
-//EG:     char *element;
-//EG:     
-//EG:     if (closeBrace == NULL) {
-//EG:       Tcl_DStringFree(&newName);      
-//EG:       Err("unmatched open-brace in file name", STk_nil);
-//EG:     }
-//EG: 
-//EG:     Tcl_DStringAppend(&newName, rem, openBrace-rem);
-//EG:     baseLength = newName.length;
-//EG:     for (p = openBrace; *p != '}'; ) {
-//EG:       element = p+1;
-//EG:       for (p = element; ((*p != '}') && (*p != ',')); p++) {}
-//EG:       Tcl_DStringAppend(&newName, element, p-element);
-//EG:       Tcl_DStringAppend(&newName, closeBrace+1, -1);
-//EG:       result = fileglob(dir, newName.string, result);
-//EG:       newName.length = baseLength;
-//EG:     }
-//EG:     goto done;
-//EG:   }
-//EG: 
-//EG:   /*
-//EG:    * Start building up the next-level name with dir plus a slash if
-//EG:    * needed to separate it from the next file name.
-//EG:    */
-//EG: 
-//EG:   Tcl_DStringAppend(&newName, dir, -1);
-//EG:   if ((dir[0] != 0) && (newName.string[newName.length-1] != '/')) {
-//EG:     Tcl_DStringAppend(&newName, SDIRSEP, 1);
-//EG:   }
-//EG:   baseLength = newName.length;
-//EG: 
-//EG:   /*
-//EG:    * If there were any pattern-matching characters, then scan through
-//EG:    * the directory to find all the matching names.
-//EG:    */
-//EG:   if (gotSpecial) {
-//EG: #ifdef _MSC_VER
-//EG:     WIN32_FIND_DATA wfd; /* VC++ support from Caleb Deupree <cdeupree@erinet.com> */
-//EG:     HANDLE handle;
-//EG:     Tcl_DString msvcname;
-//EG:     char savedChar;
-//EG:     BOOL bFound = TRUE;
-//EG: 
-//EG:     Tcl_DStringInit(&msvcname);
-//EG:     Tcl_DStringAppend(&msvcname, dirName, -1);
-//EG:     Tcl_DStringAppend(&msvcname, SDIRSEP, 1);
-//EG:     Tcl_DStringAppend(&msvcname, "*\0", -1);
-//EG:        
-//EG:     handle = FindFirstFile(Tcl_DStringValue(&msvcname), &wfd);
-//EG:     if (handle == INVALID_HANDLE_VALUE) {
-//EG:       Tcl_DStringFree(&msvcname);
-//EG:       Err("Cannot find files, error = ", STk_makestring((char *) GetLastError()));
-//EG:     }
-//EG:    
-//EG:     savedChar = *p;
-//EG:     *p = 0;
-//EG:     
-//EG:     while (bFound) {
-//EG:       /*
-//EG:        * Don't match names starting with "." unless the "." is
-//EG:        * present in the pattern.
-//EG:        */
-//EG:       if ((wfd.cFileName == '.') && (*rem != '.')) continue;
-//EG:       
-//EG:       if (Tcl_StringMatch(wfd.cFileName, rem)) {
-//EG: 	newName.length = baseLength;
-//EG: 	Tcl_DStringAppend(&newName, wfd.cFileName, -1);
-//EG: 	if (savedChar == 0)
-//EG: 	  result = Cons(STk_makestring(newName.string), result);
-//EG: 	else {
-//EG: 	  result = fileglob(newName.string, p+1, result);
-//EG: 	  if (result != TCL_OK) break;
-//EG: 	}
-//EG:       }
-//EG:       bFound = FindNextFile(handle, &wfd);
-//EG:     }
-//EG:     FindClose(handle);
-//EG:     Tcl_DStringFree(&msvcname);
-//EG:     *p = savedChar;
-//EG:     goto done;
-//EG: #else
-//EG:     DIR *d;
-//EG:     struct dirent *entryPtr;
-//EG:     char savedChar;
-//EG: 
-//EG:     d = opendir(dirName);
-//EG:     if (d == NULL) {
-//EG:       Tcl_DStringFree(&newName);
-//EG:       Err("cannot read directory", STk_makestring(dirName));
-//EG:     }
-//EG: 
-//EG:     /*
-//EG:      * Temporarily store a null into rem so that the pattern string
-//EG:      * is now null-terminated.
-//EG:      */
-//EG: 
-//EG:     savedChar = *p;
-//EG:     *p = 0;
-//EG: 
-//EG:     while (1) {
-//EG:       entryPtr = readdir(d);
-//EG:       if (entryPtr == NULL) break;
-//EG:       /*
-//EG:        * Don't match names starting with "." unless the "." is
-//EG:        * present in the pattern.
-//EG:        */
-//EG:       if ((*entryPtr->d_name == '.') && (*rem != '.')) continue;
-//EG: 
-//EG:       if (Tcl_StringMatch(entryPtr->d_name, rem)) {
-//EG: 	newName.length = baseLength;
-//EG: 	Tcl_DStringAppend(&newName, entryPtr->d_name, -1);
-//EG: 	if (savedChar == 0)
-//EG: 	  result = Cons(STk_makestring(newName.string), result);
-//EG: 	else {
-//EG: 	  result = fileglob(newName.string, p+1, result);
-//EG: 	  if (result != TCL_OK) break;
-//EG: 	}
-//EG:       }
-//EG:       
-//EG:     }
-//EG:     closedir(d);
-//EG:     *p = savedChar;
-//EG:     goto done;
-//EG: #endif
-//EG:   }
-//EG: 
-//EG:   /*
-//EG:    * The current element is a simple one with no fancy features.  Add
-//EG:    * it to the new name.  If there are more elements still to come,
-//EG:    * then recurse to process them.
-//EG:    */
-//EG:   
-//EG:   Tcl_DStringAppend(&newName, rem, p-rem);
-//EG:   if (*p != 0) {
-//EG:     result = fileglob(newName.string, p+1, result);
-//EG:     goto done;
-//EG:   }
-//EG: 
-//EG:   /*
-//EG:    * There are no more elements in the pattern.  Check to be sure the
-//EG:    * file actually exists, then add its name to the list being formed
-//EG:    * in main_interp-result.
-//EG:    */
-//EG: 
-//EG:   name = newName.string;
-//EG:   if (*name == 0) name = ".";
-//EG:   if (access(name, F_OK) != 0) goto done;
-//EG:   result = Cons(STk_makestring(name), result);
-//EG: done:
-//EG:   Tcl_DStringFree(&newName);
-//EG:   return result;
-//EG: }
-//EG: 
-#endif 
-
-SCM STk_int_expand_file_name(char *s)
-{
-  return STk_Cstring2string(s);			//FIXME
-#ifdef FIXME
-//  char expanded[2 * MAX_PATH_LENGTH], abs[2 * MAX_PATH_LENGTH];  
-//   /* Warning: absolute makes no control about path overflow. Hence the "2 *" */
-// 
-//   absolute(tilde_expand(s, expanded), abs);
-//   return STk_makestring(abs);
-#endif
-}
-
 
 #ifdef FIXME
 //EG: void STk_whence(char *exec, char *path)
@@ -473,23 +142,47 @@ SCM STk_int_expand_file_name(char *s)
  *
  ******************************************************************************/
 
-#ifdef FIXME 
-//EG: PRIMITIVE STk_expand_file_name(SCM s)
-//EG: {
-//EG:   if (NSTRINGP(s)) Err("expand-file-name: bad string", s);
-//EG:   return STk_internal_expand_file_name(CHARS(s));
-//EG: }
-//EG: 
-//EG: PRIMITIVE STk_canonical_path(SCM str)
-//EG: {
-//EG:   if (NSTRINGP(str)) Err("canonical-path: not a string", str);
-//EG: #ifdef WIN32
-//EG:   return str;
-//EG: #else
-//EG:   return STk_resolve_link(CHARS(str), 0);
-//EG: #endif
-//EG: }
-#endif
+/*
+<doc EXT expand-file-name
+ * (expand-file-name path)
+ * 
+ * @cindex tilde expansion
+ * |Expand-file-name| expands the filename given in |path| to
+ * an absolute path. 
+ * @lisp
+ *   ;; Current directory is ~eg/STklos (i.e. /users/eg/STklos)
+ *   (expand-file-name "..")            => "/users/eg"
+ *   (expand-file-name "~eg/../eg/bin") => "/users/eg/bin"
+ *   (expand-file-name "~/STklos)"      => "/users/eg/STk"
+ * @end lisp 
+doc>
+*/
+
+DEFINE_PRIMITIVE("expand-file-name", expand_fn, subr1, (SCM s))
+{
+  ENTER_PRIMITIVE(expand_fn);
+  if (!STRINGP(s)) error_bad_string(s);
+  return STk_Cstring2string(STk_expand_file_name(STRING_CHARS(s)));
+}
+
+
+/*
+<doc EXT canonical-file-name
+ * (canonical-file-name path)
+ *
+ * Expands all symbolic links in |path| and returns its canonicalized
+ * absolute path name. The resulting path does not have symbolic links. 
+ * If |path| doesn't designate a valid path name, |canonical-file-name| 
+ * returns |f|.
+doc>
+*/
+DEFINE_PRIMITIVE("canonical-file-name", canonical_path, subr1, (SCM str))
+{
+  ENTER_PRIMITIVE(canonical_path);
+
+  if (!STRINGP(str)) error_bad_string(str);
+  return STk_resolve_link(STRING_CHARS(str), 0);
+}
 
 /*
 <doc EXT getcwd
@@ -526,7 +219,7 @@ DEFINE_PRIMITIVE("chdir", chdir, subr1, (SCM s))
 
   if (!STRINGP(s)) error_bad_path(s);
   
-  if (chdir(STRING_CHARS(STk_int_expand_file_name(STRING_CHARS(s)))))
+  if (chdir(STk_expand_file_name(STRING_CHARS(s))))
     STk_error("cannot change directory to ~S", s);
  
   return STk_void;
@@ -568,7 +261,7 @@ DEFINE_PRIMITIVE("system", system, subr1, (SCM com))
  * (file-is-directory?  string)
  * (file-is-regular?    string)
  * (file-is-readable?   string)
- * (file-is-writable?   string)
+x * (file-is-writable?   string)
  * (file-is-executable? string)
  * (file-exists?        string)
  *
@@ -630,23 +323,46 @@ DEFINE_PRIMITIVE("file-exists?", file_existsp, subr1, (SCM f))
   return my_access(f, F_OK);
 }
 
-#ifdef FIXME
-//EG: PRIMITIVE STk_file_glob(SCM l, int len) /* len is unused here */
-//EG: {
-//EG:   SCM res = STk_nil;
-//EG:   char s[2*MAX_PATH_LENGTH];
-//EG:   
-//EG:   for ( ; NNULLP(l); l = CDR(l)) {
-//EG:     if (NSTRINGP(CAR(l))) Err("glob: bad string", CAR(l));
-//EG:     
-//EG:     tilde_expand(CHARS(CAR(l)), s);
-//EG: 
-//EG:     res = STk_append2(res, (ISDIRSEP(*s)) ? fileglob(SDIRSEP, s+1, STk_nil) :
-//EG: 		                            fileglob("", s, STk_nil));
-//EG:   }
-//EG:   return res;
-//EG: }
-#endif
+/*
+<doc EXT glob
+ * (glob pattern ...)
+ *
+ * |Glob| performs file name ``globbing'' in a fashion similar to the 
+ * csh shell. |Glob| returns a list of the filenames that match at least
+ * one of |pattern| arguments.  The |pattern| arguments may contain
+ * the following special characters:
+ * @itemize @bullet
+ * @item |?| Matches any single character.
+ * @item |*| Matches any sequence of zero or more characters.
+ * @item |[chars]| Matches any single character in |chars|. 
+ * If chars contains a sequence of the form |a-b| then any character 
+ * between |a| and |b| (inclusive) will match.
+ * @item  |\\| Matches the character |x|.
+ * @item |@{a,b,...@}| Matches any of the strings |a|, |b|, etc.
+ * @end itemize
+ * 
+ * As with csh, a '.' at the beginning of a file's name or just after 
+ * a '/ must be matched explicitly or with a |@{@}| construct.  
+ * In addition, all '/' characters must be matched explicitly.
+ * 
+ * If the first character in a pattern is '~' then it refers to
+ * the home directory of the user whose name follows the '~'.
+ * If the '~' is followed immediately by `/' then the value of
+ * the environment variable HOME is used.
+ *
+ * |Glob| differs from csh globbing in two ways.  First, it does not
+ * sort its result list (use the |sort| procedure if you want the list
+ * sorted).
+ * Second, |glob| only returns the names of files that actually exist; 
+ * in csh no check for existence is made unless a pattern contains a 
+ * |?|, |*|, or |@{@}| construct.
+doc>
+*/
+DEFINE_PRIMITIVE("glob", glob, vsubr, (int argc, SCM *argv))
+{
+  ENTER_PRIMITIVE(glob);
+  return STk_do_glob(argc, argv);
+}
 
 
 /*
@@ -687,6 +403,46 @@ DEFINE_PRIMITIVE("rename-file", rename_file, subr2, (SCM filename1, SCM filename
   return STk_void;
 }
  
+
+/*
+<doc EXT copy-file
+ * (copy-file string1 string2)
+ *
+ * Copies the file whose path-name is |string1| to a file whose path-name is
+ * |string2|. If the file |string2| already exists, its content prior 
+ * the call to |copy-file| is lost. The result of |copy-file| is @emph{void}.
+doc>
+*/
+DEFINE_PRIMITIVE("copy-file", copy_file, subr2, (SCM filename1, SCM filename2))
+{
+  char buff[1024];
+  int f1, f2, n;
+
+  /* Should I use sendfile on Linux here? */
+  ENTER_PRIMITIVE(copy_file);
+  
+  if (!STRINGP(filename1)) error_bad_string(filename1);
+  if (!STRINGP(filename2)) error_bad_string(filename2);
+  
+  f1 = open(STRING_CHARS(filename1), O_RDONLY);
+  f2 = open(STRING_CHARS(filename2), O_WRONLY|O_CREAT|O_TRUNC, 0666);
+  
+  if ((f1==-1) || (f2==-1)) {
+    error_cannot_copy(filename1, filename2);
+  }
+
+  while ((n = read(f1, buff, MAXBUFF)) > 0) {
+    if ((n < 0) || (write(f2, buff, n) < n)) {
+      close(f1); close(f2); 
+      error_cannot_copy(filename1, filename2);
+    }
+  }
+  
+  close(f1); close(f2);
+  return STk_void;
+}
+
+
 /*
 <doc EXT temporary-file-name
  * (temporary-file-name)
@@ -698,10 +454,24 @@ doc>
 */
 DEFINE_PRIMITIVE("temporary-file-name", tmp_file, subr0, (void))
 {
+#ifdef WIN32
   char buff[MAX_PATH_LENGTH], *s;
   
   s = tmpnam(buff);
   return s ? STk_Cstring2string(s) : STk_false;
+#else
+  static int cpt=0;
+  char buff[MAX_PATH_LENGTH];
+
+  for ( ; ; ) {
+    sprintf(buff, "/tmp/stklos%05x", cpt++);
+    if (cpt > 100000)		/* arbitrary limit to avoid infinite search */
+      return STk_false; 
+    if (access(buff, F_OK) == -1) break;
+  }
+
+  return STk_Cstring2string(buff);
+#endif
 }
 
 
@@ -722,7 +492,7 @@ DEFINE_PRIMITIVE("exit", quit, subr01, (SCM retcode))
 
   if (retcode) {
     ret = STk_integer_value(retcode);
-    if (retcode == LONG_MIN) STk_error("bad return code ~S", retcode);
+    if (ret == LONG_MIN) STk_error("bad return code ~S", retcode);
   }
 #ifdef FIXME
 //EG:  /* Execute all the terminal thunks of pending dynamic-wind */
@@ -839,7 +609,151 @@ doc>
  */
 DEFINE_PRIMITIVE("current-time", current_time, subr0, (void))
 {
-  return STk_long2integer(time(NULL));
+  return STk_ulong2integer(time(NULL));
+}
+
+/*
+<doc EXT full-current-time
+ * (full-current-time)
+ * 
+ * Returns the time of the day as a pair where 
+ * @itemize -
+ * @item the  first element is the time since the Epoch 
+ * (that is 00:00:00 UTC, January 1, 1970),  measured in seconds. 
+ * 
+ * @item the second element is the number of microseconds in the given 
+ * second.
+ * @end itemize
+doc>
+*/
+DEFINE_PRIMITIVE("full-current-time", full_current_time, subr0, (void))
+{
+  struct timeval now;
+  gettimeofday(&now, NULL); 
+
+  return STk_cons(STk_long2integer(now.tv_sec),
+		  STk_long2integer(now.tv_usec));
+}
+		  
+/*
+<doc EXT seconds->date
+ * (seconds->date sec)
+ * 
+ * Returns a keyword list for the date given by |sec| (a date based on the 
+ * Epoch). The keyed values returned are
+ * @itemize -
+ * @item second : 0 to 59 (but can be up to 61 to allow for leap seconds)
+ * @item minute : 0 to 59
+ * @item hour : 0 to 23
+ * @item day : 1 to 31
+ * @item month : 1 to 12
+ * @item year : e.g., 2002
+ * @item week-day : 0 (Sunday) to 6 (Saturday)
+ * @item year-day : 0 to 365 (365 in leap years)
+ * @item dst? : #t (daylight savings time) or #f
+ * @item time-zone-offset : the difference between Coordinated Universal Time
+ * (UTC) and local standard time in seconds.
+ * @end itemize
+ * @noindent
+ * Example:
+ * @lisp
+ * (seconds->date (current-time))   => (:second 49 :minute 32 :hour 23
+ *                                      :day 2 :month 4 :year 2002 
+ *                                      :week-day 2 :year-day 91 
+ *                                      :dst #t :time-zone-offset -3600)
+ * @end lisp
+doc>
+*/
+DEFINE_PRIMITIVE("%seconds->date", seconds2date, subr1, (SCM seconds))
+{
+  int overflow; 
+  SCM v, *vect;
+  struct tm *t;
+  time_t tt;
+
+  ENTER_PRIMITIVE(seconds2date);
+  tt = (time_t) STk_integer2int32(seconds, &overflow);
+  
+  if (overflow) error_bad_int_or_out_of_bounds(seconds);
+  
+  t        = localtime(&tt);
+
+  v 	   = STk_makevect(20, STk_false);
+  vect     = VECTOR_DATA(v);
+  
+  vect[0]  = STk_makekey("second");
+  vect[1]  = MAKE_INT(t->tm_sec);
+  vect[2]  = STk_makekey("minute");
+  vect[3]  = MAKE_INT(t->tm_min);
+  vect[4]  = STk_makekey("hour");
+  vect[5]  = MAKE_INT(t->tm_hour);
+  vect[6]  = STk_makekey("day");
+  vect[7]  = MAKE_INT(t->tm_mday);
+  vect[8]  = STk_makekey("month");
+  vect[9]  = MAKE_INT(t->tm_mon + 1);
+  vect[10] = STk_makekey("year");
+  vect[11] = MAKE_INT(1900 + t->tm_year);
+  vect[12] = STk_makekey("week-day");
+  vect[13] = MAKE_INT(t->tm_wday);
+  vect[14] = STk_makekey("year-day");
+  vect[15] = MAKE_INT(t->tm_yday);
+  vect[16] = STk_makekey("dst");
+  vect[17] = MAKE_BOOLEAN(t->tm_isdst > 0);
+  vect[18] = STk_makekey("time-zone-offset");
+  vect[19] = STk_long2integer(timezone);
+
+  return STk_vector2list(v);
+}
+
+DEFINE_PRIMITIVE("%date->seconds", date2seconds, subr1, (SCM v))
+{
+  struct tm t;
+  SCM *vect;
+  time_t n;
+
+  ENTER_PRIMITIVE(date2seconds);
+  
+  if (!(VECTORP(v) && VECTOR_SIZE(v) == 7)) STk_error("bad date vector ~S", v);
+  vect = VECTOR_DATA(v);
+
+  t.tm_sec   = STk_integer_value(vect[0]);
+  t.tm_min   = STk_integer_value(vect[1]);
+  t.tm_hour  = STk_integer_value(vect[2]);
+  t.tm_mday  = STk_integer_value(vect[3]);
+  t.tm_mon   = STk_integer_value(vect[4]) - 1;
+  t.tm_year  = STk_integer_value(vect[5]) - 1900;
+  t.tm_isdst = (vect[6] != STk_false);
+
+  n = mktime(&t);
+  if (n == (time_t)(-1)) STk_error("cannot convert date to seconds (~S)", v);
+  
+  return STk_ulong2integer((long) n);
+}
+
+
+DEFINE_PRIMITIVE("%seconds->string", date2string, subr2, (SCM fmt, SCM seconds))
+{
+  char buffer[1024];
+  struct tm *p;
+  time_t tt;  
+  int len, overflow;
+
+  ENTER_PRIMITIVE(date2string);
+  
+  tt = (time_t) STk_integer2int32(seconds, &overflow);
+  
+  if (!STRINGP(fmt)) error_bad_string(fmt);
+  if (overflow)      error_bad_int_or_out_of_bounds(seconds);
+  
+  p   = localtime(&tt);
+  len = strftime(buffer, 1023, STRING_CHARS(fmt), p);
+
+  if (len > 0)
+    return STk_Cstring2string(buffer);
+  else 
+    STk_error("buffer too short!");
+  
+  return STk_void; /* never reached */
 }
 
 
@@ -867,23 +781,50 @@ DEFINE_PRIMITIVE("running-os", running_os, subr0, (void))
 /*
 <doc EXT getenv
  * (getenv str)
+ * (getenv)
  *
  * Looks for the environment variable named |str| and returns its
  * value as a string, if it exists. Otherwise, |getenv| returns |#f|.
+ * If |getenv| is called without parameter, it returns the list of 
+ * all the environment variables accessible from the program as an 
+ * A-list.
  * @lisp
- * (getenv "SHELL")   => "/bin/zsh"
+ * (getenv "SHELL")   
+ *      => "/bin/zsh"
+ * (getenv)
+ *      => (("TERM" . "xterm") ("PATH" . "/bin:/usr/bin") ...)
  * @end lisp
 doc>
  */
-DEFINE_PRIMITIVE("getenv", getenv, subr1, (SCM str))
+static SCM build_posix_environment(char **env)
+{
+  SCM l;
+
+  for (l=STk_nil; *env; env++) {
+    char *s, *p;
+
+    s = *env; p = strchr(s, '=');
+    if (p)
+      l = STk_cons(STk_cons(STk_makestring(p-s, s), STk_Cstring2string(p+1)),l);
+  }
+  return l;
+}
+
+
+DEFINE_PRIMITIVE("getenv", getenv, subr01, (SCM str))
 {
   char *tmp;
   
   ENTER_PRIMITIVE(getenv);
-  if (!STRINGP(str)) error_bad_string(str);
-  
-  tmp = getenv(STRING_CHARS(str));
-  return tmp ? STk_Cstring2string(tmp) : STk_false;
+  if (str) {		/* One parameter: find the value of the given variable */
+    if (!STRINGP(str)) error_bad_string(str);
+    
+    tmp = getenv(STRING_CHARS(str));
+    return tmp ? STk_Cstring2string(tmp) : STk_false;
+  } else {		/* No parameter: give the complete environment */
+    extern char **environ;
+    return build_posix_environment(environ);
+  }
 }
 
 /*
@@ -926,6 +867,10 @@ int STk_init_system(void)
 {
   ADD_PRIMITIVE(clock);
   ADD_PRIMITIVE(current_time);
+  ADD_PRIMITIVE(full_current_time);
+  ADD_PRIMITIVE(seconds2date);
+  ADD_PRIMITIVE(date2seconds);
+  ADD_PRIMITIVE(date2string);
   ADD_PRIMITIVE(running_os);
   ADD_PRIMITIVE(getenv);
   ADD_PRIMITIVE(setenv);
@@ -942,8 +887,13 @@ int STk_init_system(void)
   ADD_PRIMITIVE(file_is_writablep);
   ADD_PRIMITIVE(file_is_executablep);
   ADD_PRIMITIVE(file_existsp);
+  ADD_PRIMITIVE(glob);
+  ADD_PRIMITIVE(expand_fn);
+  ADD_PRIMITIVE(canonical_path);
+
   ADD_PRIMITIVE(remove_file);
   ADD_PRIMITIVE(rename_file);
+  ADD_PRIMITIVE(copy_file);
   ADD_PRIMITIVE(tmp_file);
   ADD_PRIMITIVE(quit);
   ADD_PRIMITIVE(machine_type);
