@@ -1,7 +1,7 @@
 /*
  * v m . c				-- The STklos Virtual Machine
  * 
- * Copyright  2000-2001 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
+ * Copyright  2000-2003 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>
  * 
  * 
  * This program is free software; you can redistribute it and/or modify
@@ -21,21 +21,27 @@
  * 
  *           Author: Erick Gallesio [eg@unice.fr]
  *    Creation date:  1-Mar-2000 19:51 (eg)
- * Last file update: 10-Mar-2001 17:54 (eg)
+ * Last file update:  9-May-2003 16:13 (eg)
  */
-
-/* #define DEBUG_VM */
-
-#ifdef DEBUG_VM
-static int debug_level = 2;	/* 0 is quiet, 1, 2, ... are more verbose */
-#endif
-
-
 #include "stklos.h"
 #include "object.h"
 #include "vm.h"
 #include "vm-instr.h"
 
+/* #define DEBUG_VM */
+/* #define STAT_VM  */
+
+
+#ifdef STAT_VM 
+#  define DEBUG_VM
+static int couple_instr[NB_VM_INSTR][NB_VM_INSTR];
+static int cpt_inst[NB_VM_INSTR];
+#endif
+
+#ifdef DEBUG_VM
+static int debug_level = 0;	/* 0 is quiet, 1, 2, ... are more verbose */
+#endif
+
 
 #if defined(__GNUC__) && !defined(DEBUG_VM)
    /* Use computed gotos to have better performances */
@@ -56,6 +62,8 @@ static int debug_level = 2;	/* 0 is quiet, 1, 2, ... are more verbose */
 #endif
 
 
+
+
 #define MY_SETJMP(jb) 		(jb.blocked = get_signal_mask(), setjmp(jb.j))
 #define MY_LONGJMP(jb, val)	(longjmp((jb).j, val))
 
@@ -112,7 +120,7 @@ static SCM *handlers;		/* Exceptions handlers		*/
 
 static SCM r1, r2;			/* general registers		 */
 static SCM val2, val3, val4, val5;	/* registers for multiple values */
-static SCM valc;			/* # of multiple values 	 */
+static int valc;			/* # of multiple values 	 */
 
 static jbuf *top_jmp_buf = NULL;  
 
@@ -219,50 +227,6 @@ static jbuf *top_jmp_buf = NULL;
 }
 
 
-
-#define DO_APPLY() {   	       	       	       	       	       	        \
-  /* look at last argument */						\
-  l   = sp[nargs-1];							\
-  len = STk_int_length(l);						\
-									\
-  if (len < 0) 								\
-    STk_error("last argument ~S is not a list", l);			\
-  else if (len == 0) {							\
-    nargs -= 1;								\
-  } else {								\
-    len -= 1;								\
-    if (len > 0) { 							\
-      /* Move the elements which are before the last one toward */	\
-      /* the top of stack */						\
-      memcpy(sp-len, sp, (nargs-1) * sizeof(SCM));			\
-      sp = sp-len;							\
-    }									\
-    									\
-    /* Unfold the last argument in place */				\
-    tmp = &sp[nargs-1];							\
-    while (!NULLP(l)) {							\
-      *tmp++ = CAR(l);							\
-      l      = CDR(l);							\
-    }									\
-    nargs += len;							\
-  }									\
-}
-
-#define DO_NEXT_METHOD() {               	\
-  SCM l;					\
-  int i, n;					\
-  						\
-  val = STk_do_next_method(val, nargs, sp, &l);	\
-  n   = STk_int_length(l);			\
-  sp += nargs - n; 				\
-  for (i = 0; i < n; i++) {			\
-    sp[i] = CAR(l);				\
-    l = CDR(l);					\
-  }						\
-  nargs = n;					\
-}
-
-
 /* 
  * 		         M i s c .
  */
@@ -299,8 +263,7 @@ static int   checked_globals_used = 0;
 }
 
 
-static void run(void);
-static void run_vm(STk_instr *code, SCM *constants, SCM envt, int save);
+static void run_vm(STk_instr *code, SCM *constants, SCM envt);
 
 
 /*===========================================================================*\
@@ -313,111 +276,24 @@ static void run_vm(STk_instr *code, SCM *constants, SCM envt, int save);
 void STk_print_vm_registers(char *msg, STk_instr *code)
 {
   if (IS_IN_STACKP(env))
-    STk_debug("%s VAL=~S PC=%d SP=%d FP=%d CST=%x ENV=%x (%d)", 
-	      msg, val, pc-code, sp-stack, fp-stack, constants, env, 
-	      (SCM*)env-stack);
+    STk_fprintf(STk_stderr, "%s VAL=~S PC=%d SP=%d FP=%d CST=%x ENV=%x (%d)\n", 
+		msg, val, pc-code, sp-stack, fp-stack, constants, env, 
+		(SCM*)env-stack);
   else
-    STk_debug("%s VAL=~S PC=%d SP=%d FP=%d CST=%x ENV=%x (%d)", 
-	      msg, val, pc-code, sp-stack, fp-stack, constants, env, 
-	      (SCM*)env-stack);
+    STk_fprintf(STk_stderr, "%s VAL=~S PC=%d SP=%d FP=%d CST=%x ENV=%x (%d)", 
+		msg, val, pc-code, sp-stack, fp-stack, constants, env, 
+		(SCM*)env-stack);
 }
 
 #endif
 
 
-#ifdef _FIXME_DEBUG_VM
-static int use_debugger = 1;
-static int breakpoint = -1;
-
-
-static void vm_debug(STk_instr *code, short byteop)
-{
-  int c, prev='\n';
-# define DEFINE_NAME_TABLE
-# include "vm-instr.h"
-
-  if (!use_debugger) return; 
-
-  /* Print the current instruction and various informations */
-  fprintf(stderr, "%03d: %20s  sp=%d fp=%d env=%x", pc-code-1, 
-	  name_table[(int)byteop], sp-stack, fp-stack, (int) env);
-  if (IS_IN_STACKP(env)) fprintf(stderr, " (%d)", (SCM*)env-stack);
-  fprintf(stderr, "\n");
-
-
-  /* See if we have a breakpoint */
-  if (breakpoint > 0) {
-    if (breakpoint == pc-code-1) {
-      /* We are on the breakpoint */
-      breakpoint = -1;
-    }
-    else {
-      return;
-    }
-  }
-
-  for ( ; ; ) {
-    switch (c = getchar()) {
-      case 'r':
-	STk_print_vm_registers();
-	break;
-      case 'b':
-	printf("Enter the value of PC "); fflush(stdout);
-	scanf("%d", &breakpoint);
-	return;
-      case 'f': {
-	int j, i = sp-stack;
-	SCM *p; 
-
-	for (p = sp; p < fp; p++) {
-	  if (p == env) {
-	    STk_debug("%d:     length: %x", i++, FRAME_LENGTH(fp));
-	    STk_debug("%d:       next: %x", i++, FRAME_NEXT(fp));
-	    STk_debug("%d:      owner: ~S", i++, FRAME_OWNER(fp));
-	    p += 3;
-	    for (j=0; j < FRAME_LENGTH(fp); j++, p++)
-	      STk_debug("%d            : ~S", i++, FRAME_LOCAL(fp, j));
-	  } else {
-	    STk_debug("%d          ~S", i++, *p++);
-	  }
-	}
-	STk_debug("%d      save env: %x",i++, (SCM*) ACT_SAVE_ENV(fp)-stack);
-	STk_debug("%d       save pc: %d",i++, (STk_instr*) ACT_SAVE_PC(fp)-code);
-	STk_debug("%d      save cst: %x",i++, (SCM*) ACT_SAVE_CST(fp));
-	STk_debug("%d       save fp: %d",i++, (SCM*) ACT_SAVE_FP(fp)-stack);
-	break;
-      }
-      case 'q': 
-	use_debugger = 0; 
-	/* no break */
-      case '\n': if (prev != '\n') break;
-      case 'c':
-	return;
-      default:
-	STk_debug("h    -- help");
-	STk_debug("r    -- registers");
-	STk_debug("f    -- current frame");
-	STk_debug("c    -- continue");
-	STk_debug("b    -- breakpoint");
-	STk_debug("q    -- quit");
-	break;
-    }
-    prev = c;
-  }
-}
-#endif
-
-
-
 static Inline SCM listify_top(int n)
 {
-  SCM res = STk_nil;
-  int len = n;
+  SCM *p, res = STk_nil;
 
-  while (n--) {
-    res = STk_cons(sp[n], res);
-  }
-  sp += len;
+  for (p = sp, sp+=n; p < sp; p++)
+    res = STk_cons(*p, res);
   return res;
 }
 
@@ -460,14 +336,8 @@ static Inline short adjust_arity(SCM func, short nargs)
       else { /* Make a list from the arguments which are on the stack. */
 	SCM res = STk_nil;
 	
-	while (nargs > min_arity) 
-	  res = STk_cons(sp[--nargs], res);
-
-	/* Place res in the variable argument. Note that in the special case
-	 * where nargs = min_arity, NIL is stored in the special place-holder 
-	 * in the activation record.
-	 */
-	sp[min_arity] = res;
+	while (nargs-- > min_arity) res = STk_cons(pop(), res);
+	push(res);
       }
       return -arity;
     }
@@ -518,23 +388,46 @@ static int add_global(SCM *ref)
  * @end lisp
 doc>
  */
-DEFINE_PRIMITIVE("apply", scheme_apply, apply, (int argc, SCM *argv))
+DEFINE_PRIMITIVE("apply", scheme_apply, apply, (int argc))
 {
-  SCM l, func, *tmp;
+  SCM l, func, *tmp, *argv;
   int len, nargs;
 
   ENTER_PRIMITIVE(scheme_apply);
   if (argc == 0) STk_error("no function given");
-  
-  func  = pop();
-  nargs = argc-1;
-  
+
+  nargs = argc - 1;
+  argv  = sp + nargs;
+  func  = *argv;
+
   if (nargs > 0) {
-    DO_APPLY();
+    /* look at last argument */
+    l   = *sp;
+    len = STk_int_length(l);
+
+    if (len < 0)
+      STk_error("last argument ~S is not a list", l);
+    else {
+      /* move all the arguments, except the last one, one cell lower in the 
+       * stack (i.e. overwrite the function to call) */
+      for (tmp = argv-1; tmp > sp; tmp--)
+	*(tmp+1) = *tmp;
+
+      sp = tmp + 2;
+      if (len != 0) {
+	/* Unfold the last argument in place */
+	while (!NULLP(l)) {
+	  push(CAR(l));
+	  l = CDR(l);
+	}
+      }
+      nargs += len-1;
+    }
   }
+
   /* Place the function to apply and its number of arguments in R1 and R2 */
   r1 = func;
-  r2 = (SCM) nargs;
+  r2 = AS_SCM(nargs);
 
   return STk_apply_call;
 }
@@ -550,26 +443,31 @@ apply_proc:
 
     case tc_instance: {
       if (PUREGENERICP(val)) {
-	SCM methods = STk_compute_applicable_methods(val, nargs, sp, FALSE);
+	SCM *argv = sp+nargs-1;
+	SCM methods, nm; 
 	
 	/* methods is the list of applicable methods. Apply the first
 	 * one with the tail of the list as first parameter
 	 * (next-method). If methods is STk_nil, that's because the
 	 * no-applicable-method triggered didn't yield an error.  
 	 */
+	methods = STk_compute_applicable_methods(val, nargs, argv, FALSE);
 	if (NULLP(methods)) { val = STk_void; return; }
 	
-	/* Push next method on stack */
-	push(STk_make_next_method(val, nargs, sp, methods));
-	val    = INST_SLOT(CAR(methods), S_procedure);
-	nargs += 1;
+	/* Place the procedure of the first method in the VAL register and
+	 * store the next method in the ``next-method'' variable. 
+	 */
+	nm   = STk_make_next_method(val, nargs, argv, methods);
+	val  = INST_SLOT(CAR(methods), S_procedure);
+	SET_NEXT_METHOD(val, nm);
 	/* NO BREAK */
       } else {
-	SCM gf, args; 
-	
+	SCM gf, args;
+
 	/* Use the MOP and do the call (apply-generic gf args) */
-	gf = val; args = listify_top(nargs);
-	push(args); push(gf);
+	args = listify_top(nargs);
+	push(val); 
+	push(args);
 	val = STk_lookup(STk_intern("apply-generic"), STk_current_module, 
 			 &gf, FALSE);
 	nargs = 2;
@@ -608,14 +506,44 @@ apply_proc:
     }
 
     case tc_next_method: {
-      DO_NEXT_METHOD();
+      SCM methods, nm, *argv, proc;
+      int i;
+      
+      methods = NXT_MTHD_METHODS(val);
+      
+      if (nargs == 0) {
+	/* no argument given, place the ones of the original call on top of stack */
+	nargs = NXT_MTHD_ARGC(val);
+	argv  = NXT_MTHD_ARGV(val);
+
+	for (i = 0; i < nargs; i++) 
+	  push(argv[i]);
+      }
+      
+      argv  = sp+nargs-1;
+
+      if (NULLP(methods)) {
+	/* Do the call (no-next-method gf method args) */
+	argv = listify_top(nargs);
+	push(NXT_MTHD_GF(val));
+	push(NXT_MTHD_METHOD(val));
+	push(argv);
+	nargs = 3;
+	val   = STk_lookup(STk_intern("no-next-method"), STk_current_module, 
+			   &proc, FALSE);
+      } else {
+	/* Call the next method after creating a new next-method */
+	nm   = STk_make_next_method(val, nargs, argv, methods);
+	val  = INST_SLOT(CAR(methods), S_procedure);
+	SET_NEXT_METHOD(val, nm);
+      }
       goto apply_proc;
     }
 
     case tc_apply: {
       /* Call the function in the call frame of apply */
-      STk_scheme_apply(nargs, sp);
-      nargs = (short) r2; 
+      STk_scheme_apply(nargs);
+      nargs = (short) AS_LONG(r2); 
       val   = r1; 
       goto apply_proc;
     }
@@ -627,16 +555,16 @@ apply_proc:
       if (nargs == 1) { val = PRIMITIVE_FUNC(val)(sp[0]);		 break;}
       goto error_invoke;
     case tc_subr2:
-      if (nargs == 2) { val = PRIMITIVE_FUNC(val)(sp[0], sp[1]);	 break;}
+      if (nargs == 2) { val = PRIMITIVE_FUNC(val)(sp[1], sp[0]);	 break;}
       goto error_invoke;
     case tc_subr3:
-      if (nargs == 3) { val = PRIMITIVE_FUNC(val)(sp[0], sp[1], sp[2]);	 break;}
+      if (nargs == 3) { val = PRIMITIVE_FUNC(val)(sp[2], sp[1], sp[0]);	 break;}
       goto error_invoke;
     case tc_subr4:
-      if (nargs==4) { val = PRIMITIVE_FUNC(val)(sp[0],sp[1],sp[2], sp[3]); break;}
+      if (nargs==4) { val = PRIMITIVE_FUNC(val)(sp[3],sp[2],sp[1], sp[0]); break;}
       goto error_invoke;
     case tc_subr5:
-      if (nargs==5) { val = PRIMITIVE_FUNC(val)(sp[0],sp[1],sp[2],sp[3],sp[4]);
+      if (nargs==5) { val = PRIMITIVE_FUNC(val)(sp[4],sp[3],sp[2],sp[1],sp[0]);
 									 break;}
       goto error_invoke;
 
@@ -646,14 +574,14 @@ apply_proc:
       goto error_invoke;
     case tc_subr12:
       if (nargs == 1) {val = PRIMITIVE_FUNC(val)(sp[0], (SCM) NULL);	 break;}
-      if (nargs == 2) {val = PRIMITIVE_FUNC(val)(sp[0], sp[1]);		 break;}
+      if (nargs == 2) {val = PRIMITIVE_FUNC(val)(sp[1], sp[0]);		 break;}
       goto error_invoke;
     case tc_subr23:
-      if (nargs == 2) { val = PRIMITIVE_FUNC(val)(sp[0],sp[1],(SCM)NULL);break;}
-      if (nargs == 3) { val = PRIMITIVE_FUNC(val)(sp[0],sp[1],sp[2]);	 break;}
+      if (nargs == 2) { val = PRIMITIVE_FUNC(val)(sp[1],sp[0],(SCM)NULL);break;}
+      if (nargs == 3) { val = PRIMITIVE_FUNC(val)(sp[2],sp[1],sp[0]);	 break;}
       goto error_invoke;
     case tc_vsubr:
-      val = PRIMITIVE_FUNC(val)(nargs, sp);				 break;
+      val = PRIMITIVE_FUNC(val)(nargs, sp+nargs-1);			 break;
 
     default: 
       STk_error("bad function ~S. Cannot be applied", val);
@@ -678,7 +606,7 @@ error_invoke:
  * function: 
  *    STk_C_apply(STk_cons, 2, MAKE_INT(1), MAKE_INT(2)) => (1 . 2)
  * If nargs is < 0, we have something similar to an "execv fucntion
- *    STk_C_apply(STk_cons, -2, Argv)
+ *    STk_C_apply(...STk_intern("cons")..., -2, Argv)
  * where Argv[0] == MAKE_INT(1) and Argv[1] == MAKE_INT(2) ==> (1 . 2)
  *
 \*===========================================================================*/
@@ -690,6 +618,7 @@ SCM STk_C_apply(SCM func, int nargs, ...)
   int i;
 
   va_start(ap, nargs);
+  //  sp -= VM_STATE_SIZE;
   SAVE_VM_STATE();				    /* Save the VM regs */
   PREP_CALL();					    /* PREPARE_CALL */
 
@@ -698,18 +627,21 @@ SCM STk_C_apply(SCM func, int nargs, ...)
     SCM *argv = va_arg(ap, SCM*);
 
     nargs = -nargs;
-    sp   -= nargs;
-    for (i = 0; i < nargs; i++) sp[i] = argv[i];
+    
+    for (i = 0; i < nargs; i++) push(*argv++);
   } else {
     /* We have nargs SCM parameters to read */
-    sp -= nargs;
-    for (i = 0; i < nargs; i++) sp[i] = va_arg(ap, SCM);
+    for (i = 0; i < nargs; i++) push(va_arg(ap, SCM));
   }
 
   val     = func;				    /* Store fun in VAL */
   code[1] = (short) nargs;			    /* Patch # of args  */
 
-  run_vm(code, constants, env, 0); 
+  //  STk_debug("STk_C_apply: entre pile %d. La pile est en %d",
+  //	    old_sp - stack, sp - stack);
+
+  //  sp  = old_sp;
+  run_vm(code, constants, env); 
   FULL_RESTORE_VM_STATE(sp);
 
   return val;
@@ -737,7 +669,7 @@ DEFINE_PRIMITIVE("%execute", execute, subr23, (SCM code, SCM consts, SCM envt))
     *p++ = (STk_instr) STk_integer_value(VECTOR_DATA(code)[i]);
 
   SAVE_VM_STATE();
-  run_vm(vinstr, VECTOR_DATA(consts), envt, 1);
+  run_vm(vinstr, VECTOR_DATA(consts), envt);
   FULL_RESTORE_VM_STATE(sp);
 
   return val;
@@ -764,19 +696,19 @@ doc>
 DEFINE_PRIMITIVE("values", values, vsubr, (int argc, SCM *argv))
 {
   switch (argc) {
-    case 0:  val = STk_void; break;
-    case 1:  val = argv[0]; break;
-    case 2:  val = argv[0]; val2 = argv[1]; break;
-    case 3:  val = argv[0]; val2 = argv[1]; val3 = argv[2]; break;
-    case 4:  val = argv[0]; val2 = argv[1]; val3 = argv[2]; val4 = argv[3]; break;
-    case 5:  val = argv[0]; val2 = argv[1]; val3 = argv[2]; val4 = argv[3];
-      	     val5= argv[4]; break;
+    case 0: val = STk_void; break;
+    case 1: val = argv[0]; break;
+    case 2: val = argv[0]; val2 = argv[-1]; break;
+    case 3: val = argv[0]; val2 = argv[-1]; val3 = argv[-2]; break;
+    case 4: val = argv[0]; val2 = argv[-1]; val3 = argv[-2]; val4 = argv[-3]; break;
+    case 5: val = argv[0]; val2 = argv[-1]; val3 = argv[-2]; val4 = argv[-3];
+      	    val5= argv[-4]; break;
     default: { /* More than 5 values. Use a vector */
 	       int i;
 	       
-	       val  = argv[0]; /*as ussal */
+	       val  = argv[0]; /*as usual */
 	       val2 = STk_makevect(argc, (SCM) NULL);
-	       for (i = 0; i < argc; i++) VECTOR_DATA(val2)[i] = argv[i];
+	       for (i = 0; i < argc; i++) VECTOR_DATA(val2)[i] = *argv--;
 	     }
   }
   /* Retain in valc the number of values */
@@ -820,7 +752,7 @@ DEFINE_PRIMITIVE("call-with-values", call_with_values, subr2, (SCM prod, SCM con
     case 3:  return STk_C_apply(con, 3, val, val2, val3);
     case 4:  return STk_C_apply(con, 4, val, val2, val3, val4);
     case 5:  return STk_C_apply(con, 5, val, val2, val3, val4, val5);
-    default: return STk_C_apply(con, -valc, VECTOR_DATA(val2));
+    default: return STk_C_apply(con, -tmp, VECTOR_DATA(val2));
   }
 }
 
@@ -834,13 +766,45 @@ SCM STk_2_values(SCM v1, SCM v2)
 
 
 #ifdef DEBUG_VM
-DEFINE_PRIMITIVE("%vm", set_vm_debug, subr1, (SCM v))
+#  ifdef STAT_VM
+#    define DEFINE_NAME_TABLE
+#    include "vm-instr.h"
+
+static void dump_couple_instr(void)
+{
+  int i, j;
+  FILE *dump;
+
+  dump = fopen("/tmp/dump.out", "w");
+  fprintf(dump, "[\n");
+
+  for (i = NOP; i < NB_VM_INSTR; i++) {
+    fprintf(dump, "((%s %d) ", name_table[i], cpt_inst[i]);
+    for (j = NOP; j < NB_VM_INSTR; j++)
+      fprintf(dump, "(%s %4d) ", name_table[j], couple_instr[i][j]);
+    fprintf(dump, ")\n");
+  }
+  fprintf(dump, "\n]\n");
+}
+# endif
+
+DEFINE_PRIMITIVE("%vm", set_vm_debug, vsubr, (int argc, SCM *argv))
 {
   /* 
    * This function is just a placeholder for debugging the VM. It's body is 
    * changed depending of the current bug to track 
    */
-  
+  printf("sp=%p\n", sp);
+  {
+    SCM Argv[2];
+    
+    Argv[0] = MAKE_INT(1);
+    Argv[1] = MAKE_INT(2);
+    STk_debug("On va appeler");
+    STk_debug("RES= ~S", STk_C_apply(argv[0], -2, &Argv));
+  }
+
+  printf("sp=%p\n", sp);
   return STk_void;
 }
 #endif 
@@ -852,22 +816,51 @@ DEFINE_PRIMITIVE("%vm", set_vm_debug, subr1, (SCM v))
  *
 \*===========================================================================*/
 
-static void run(void)
+static void run_vm(STk_instr *code, SCM *consts, SCM envt)
 {
+  jbuf jb, *old_jb;
   short offset, nargs=0;
-#ifdef USE_COMPUTED_GOTO
-#  define  DEFINE_JUMP_TABLE
+#if defined(USE_COMPUTED_GOTO) 
+#  define DEFINE_JUMP_TABLE
 #  include "vm-instr.h"
-  NEXT;
 #else
-  static STk_instr *code_base = NULL;
-  short byteop;
-#  ifdef DEBUG_VM
+   short byteop;
+#endif
+#if defined(DEBUG_VM)
 #    define DEFINE_NAME_TABLE
 #    include "vm-instr.h"
-   if (!code_base) code_base = pc;
-#  endif
-   
+  static STk_instr *code_base = NULL;
+#endif
+#if defined(STAT_VM)
+  static short previous_op = NOP;
+#endif
+
+  /*
+  **
+  ** Initialize the VM 
+  **
+  */
+  old_jb      = top_jmp_buf;
+  top_jmp_buf = &jb;
+
+  constants   = consts;
+  env         = envt;
+  pc	      = code;
+
+  if (MY_SETJMP(jb)) {
+    /* We come back from an error. */
+    set_signal_mask(jb.blocked);
+  }
+
+
+  /*
+  **
+  ** Run the VM 
+  **
+  */  
+#if defined(USE_COMPUTED_GOTO)
+  NEXT;
+#else
   for ( ; ; ) {		
     /* Execution loop */
     byteop = fetch_next();
@@ -877,6 +870,11 @@ static void run(void)
 	      pc - 1,
 	      pc-code_base-1, 
 	      name_table[(int)byteop], sp-stack, fp-stack, (int) env);
+#    ifdef STAT_VM
+    couple_instr[previous_op][byteop]++;
+    cpt_inst[byteop]++;
+    previous_op = byteop;
+#    endif
 #  endif
     switch (byteop) {
 #endif /*  USE_COMPUTED_GOTO */
@@ -893,9 +891,20 @@ CASE(IM_ZERO)   { val = MAKE_INT(0);     NEXT;}
 CASE(IM_ONE)    { val = MAKE_INT(1);     NEXT;}
 CASE(IM_VOID)   { val = STk_void; NEXT;}
 
+CASE(SMALL_INT) { val = MAKE_INT(fetch_next());		NEXT;}
+CASE(CONSTANT)  { val = fetch_const();	        	NEXT;}
 
-CASE(SMALL_INT) { val = MAKE_INT(fetch_next());	NEXT;}
-CASE(CONSTANT)  { val = fetch_const();	        NEXT;}
+CASE(FALSE_PUSH)  { push(val = STk_false);       NEXT;}
+CASE(TRUE_PUSH)   { push(val = STk_true);        NEXT;}
+CASE(NIL_PUSH)    { push(val = STk_nil);         NEXT;}
+CASE(MINUS1_PUSH) { push(val = MAKE_INT(-1));    NEXT;}
+CASE(ZERO_PUSH)   { push(val = MAKE_INT( 0));    NEXT;}
+CASE(ONE_PUSH)    { push(val = MAKE_INT(+1));    NEXT;}
+CASE(VOID_PUSH)   { push(val = STk_void);        NEXT;}
+
+
+CASE(INT_PUSH)      { push(val=MAKE_INT(fetch_next())) ; NEXT; }
+CASE(CONSTANT_PUSH) { push(val=fetch_const()); 		 NEXT; }
 
 
 CASE(GLOBAL_REF) {
@@ -909,6 +918,26 @@ CASE(GLOBAL_REF) {
 }
 CASE(UGLOBAL_REF) { /* Never produced by compiler */ val = fetch_global(); NEXT; }
 
+CASE(GREF_INVOKE) {
+  SCM ref;
+
+  val   = STk_lookup(fetch_const(), env, &ref, TRUE);
+  nargs = fetch_next();
+  /* patch the code for optimize next accesses (pc[-1] is already equal to nargs)*/
+  pc[-3] = UGREF_INVOKE;
+  pc[-2] = add_global(&CDR(ref));
+
+  /*and now invoke */
+  funcall(nargs, FALSE);
+  NEXT;
+}
+
+CASE(UGREF_INVOKE) { /* Never produced by compiler */ 
+  val   = fetch_global();
+  nargs = fetch_next();
+  funcall(nargs, FALSE);
+  NEXT; 
+}
 
 CASE(LOCAL_REF0) { val = FRAME_LOCAL(env, 0); NEXT;}
 CASE(LOCAL_REF1) { val = FRAME_LOCAL(env, 1); NEXT;}
@@ -944,7 +973,6 @@ CASE(UGLOBAL_SET) { /* Never produced by compiler */
   fetch_global() = val; val = STk_void; NEXT;
 }
 
-
 CASE(LOCAL_SET0) { FRAME_LOCAL(env, 0)         = val; val = STk_void; NEXT;}
 CASE(LOCAL_SET1) { FRAME_LOCAL(env, 1)         = val; val = STk_void; NEXT;}
 CASE(LOCAL_SET2) { FRAME_LOCAL(env, 2)         = val; val = STk_void; NEXT;}
@@ -977,6 +1005,34 @@ CASE(JUMP_TRUE) {
   NEXT;
 }
 
+CASE(JUMP_NUMDIFF) {
+  offset = fetch_next(); if (!STk_numeq2(pop(), val)) pc += offset; NEXT;
+}
+CASE(JUMP_NUMEQ) {
+  offset = fetch_next(); if (STk_numeq2(pop(), val)) pc += offset; NEXT;
+}
+CASE(JUMP_NUMLT) {
+  offset = fetch_next(); if (STk_numlt2(pop(), val)) pc += offset; NEXT;
+}
+CASE(JUMP_NUMLE) {
+  offset = fetch_next(); if (STk_numle2(pop(), val)) pc += offset; NEXT;
+}
+CASE(JUMP_NUMGT) {
+  offset = fetch_next(); if (STk_numgt2(pop(), val)) pc += offset; NEXT;
+}
+CASE(JUMP_NUMGE) {
+  offset = fetch_next(); if (STk_numge2(pop(), val)) pc += offset; NEXT;
+}
+CASE(JUMP_NOT_EQ) {
+  offset = fetch_next(); if ((pop() != val)) pc += offset; NEXT;
+}
+CASE(JUMP_NOT_EQV) {
+  offset = fetch_next(); if ((STk_eqv(pop(), val) == STk_false)) pc += offset; NEXT;
+}
+CASE(JUMP_NOT_EQUAL) {
+  offset = fetch_next(); if ((STk_equal(pop(), val)==STk_false)) pc += offset; NEXT;
+}
+
 
 CASE(DEFINE_SYMBOL) {
   SCM var = fetch_const();
@@ -1015,6 +1071,7 @@ CASE(PREPARE_CALL) { PREP_CALL(); NEXT; }
 CASE(RETURN) 	   { RET_CALL();  NEXT; }
 CASE(INVOKE)       {
   nargs = fetch_next();
+invoke:
   funcall(nargs, FALSE);
   NEXT;
 }
@@ -1025,6 +1082,17 @@ CASE(TAIL_INVOKE) {
   NEXT;
 }
 
+CASE(APPLY_C_FUNC) {	// FIXME: a virer
+  SAVE_VM_STATE();
+  //  PREP_CALL();
+  nargs = fetch_next();
+  sp -= ACT_RECORD_SIZE + nargs;
+  // STk_debug("Avant le funcall sp = %d", sp - stack);
+  /* arguments are already on the stack */
+  goto invoke;
+}
+
+
 
 CASE(ENTER_LET_STAR) { sp += (int) pc[1] - 1; /* NO NEXT */ }
 CASE(ENTER_LET) {
@@ -1070,7 +1138,7 @@ CASE(ENTER_TAIL_LET) {
 
 
 CASE(PUSH_HANDLER) {
-  SCM offset = fetch_next();
+  int offset = fetch_next();
 
   /* place the value in val on the stack as well as the value of handlers */
   if (STk_procedurep(val) == STk_false)
@@ -1101,7 +1169,9 @@ CASE(MAKE_EXPANDER) {
 }
 
 CASE(END_OF_CODE) { 
-  return; 
+  /* Reset the jumbuf we have on entrance */
+  top_jmp_buf = old_jb;
+  return;
 }
 
 CASE(EVAL_CODE) {
@@ -1109,13 +1179,12 @@ CASE(EVAL_CODE) {
 }
 
 CASE(EXEC_HANDLER) {
-  SCM value, location;
+  SCM kind, location;
 
-  /* (%%exec kind location value) => value @sp[1], location @sp[0], kind in val */
+  /* (%%exec kind location value) => value in val, location @sp[0], kind @sp[1] */
   location  = pop();
-  value     = pop();
-  
-  STk_execute_current_handler(val, location, value);
+  kind      = pop();
+  STk_execute_current_handler(kind, location, val);
 }
  
   
@@ -1126,41 +1195,49 @@ CASE(EXEC_HANDLER) {
  * 			     I n l i n e d   F u n c t i o n s
  *
  ******************************************************************************/
-CASE(IN_ADD2)   { val = STk_add2(val, pop());		           NEXT;}
-CASE(IN_SUB2)   { val = STk_sub2(val, pop());		           NEXT;}
-CASE(IN_MUL2)   { val = STk_mul2(val, pop());		           NEXT;}
-CASE(IN_DIV2)   { val = STk_div2(val, pop());		           NEXT;}
+#define SCHEME_NOT(x) (((x) == STk_false) ? STk_true: STk_false)
 
-CASE(IN_NUMEQ)  { val = MAKE_BOOLEAN(STk_numeq2(val, pop()));      NEXT;}
-CASE(IN_NUMLT)  { val = MAKE_BOOLEAN(STk_numlt2(val, pop()));      NEXT;}
-CASE(IN_NUMGT)  { val = MAKE_BOOLEAN(STk_numgt2(val, pop()));      NEXT;}
-CASE(IN_NUMLE)  { val = MAKE_BOOLEAN(STk_numle2(val, pop()));      NEXT;}
-CASE(IN_NUMGE)  { val = MAKE_BOOLEAN(STk_numge2(val, pop()));      NEXT;}
+
+CASE(IN_ADD2)   { val = STk_add2(pop(), val);		           NEXT;}
+CASE(IN_SUB2)   { val = STk_sub2(pop(), val);		           NEXT;}
+CASE(IN_MUL2)   { val = STk_mul2(pop(), val);		           NEXT;}
+CASE(IN_DIV2)   { val = STk_div2(pop(), val);		           NEXT;}
+
+CASE(IN_NUMEQ)  { val = MAKE_BOOLEAN(STk_numeq2(pop(), val));      NEXT;}
+CASE(IN_NUMDIFF){ val = MAKE_BOOLEAN(!STk_numeq2(pop(), val));     NEXT;}
+CASE(IN_NUMLT)  { val = MAKE_BOOLEAN(STk_numlt2(pop(), val));      NEXT;}
+CASE(IN_NUMGT)  { val = MAKE_BOOLEAN(STk_numgt2(pop(), val));      NEXT;}
+CASE(IN_NUMLE)  { val = MAKE_BOOLEAN(STk_numle2(pop(), val));      NEXT;}
+CASE(IN_NUMGE)  { val = MAKE_BOOLEAN(STk_numge2(pop(), val));      NEXT;}
 
 CASE(IN_INCR)   { val = STk_add2(val, MAKE_INT(1));		   NEXT;}
 CASE(IN_DECR)   { val = STk_sub2(val, MAKE_INT(1));		   NEXT;}
 
-CASE(IN_CONS)   { val = STk_cons(val, pop());		           NEXT;}
+CASE(IN_CONS)   { val = STk_cons(pop(), val);		           NEXT;}
 CASE(IN_CAR)    { val = STk_car(val);			           NEXT;}
 CASE(IN_CDR)    { val = STk_cdr(val);			           NEXT;}
 CASE(IN_NULLP)  { val = MAKE_BOOLEAN(val == STk_nil);	           NEXT;}
 CASE(IN_LIST)   { val = listify_top(fetch_next());	           NEXT;}
-CASE(IN_NOT)    { val = (val==STk_false) ? STk_true: STk_false;    NEXT;}
+CASE(IN_NOT)    { val = SCHEME_NOT(val); 			   NEXT;}
 
-CASE(IN_EQUAL)  { val = STk_equal(val, pop());			   NEXT;}
-CASE(IN_EQV)    { val = STk_eqv(val, pop());			   NEXT;}
-CASE(IN_EQ)     { val = MAKE_BOOLEAN(val == pop());		   NEXT;}
+CASE(IN_EQUAL)  { val = STk_equal(pop(), val);			   NEXT;}
+CASE(IN_EQV)    { val = STk_eqv(pop(), val);			   NEXT;}
+CASE(IN_EQ)     { val = MAKE_BOOLEAN(pop() == val);		   NEXT;}
 
-CASE(IN_VREF)   { val = STk_vector_ref(val, pop()); 		   NEXT;}
-CASE(IN_SREF)   { val = STk_string_ref(val, pop()); 		   NEXT;}
-CASE(IN_VSET)   { 
-  SCM arg2 = pop();
-  val = STk_vector_set(val, arg2, pop());
+CASE(IN_NOT_EQUAL) { val = SCHEME_NOT(STk_equal(pop(), val)); 	   NEXT; }
+CASE(IN_NOT_EQV)   { val = SCHEME_NOT(STk_eqv(pop(), val)); 	   NEXT; }
+CASE(IN_NOT_EQ)    { val = MAKE_BOOLEAN(pop() != val);		   NEXT; }
+
+CASE(IN_VREF)   { val = STk_vector_ref(pop(), val); 		   NEXT;}
+CASE(IN_SREF)   { val = STk_string_ref(pop(), val); 		   NEXT;}
+CASE(IN_VSET)   {
+  SCM index = pop();
+  val = STk_vector_set(pop(), index, val);
   NEXT;
 }
 CASE(IN_SSET)   { 
-  SCM arg2 = pop();
-  val = STk_string_set(val, arg2, pop());
+  SCM index = pop();
+  val = STk_string_set(pop(), index, val);
   NEXT;
 }
 
@@ -1179,29 +1256,29 @@ CASE(IN_APPLY)   {
 }
 
 
-static void run_vm(STk_instr *code, SCM *consts, SCM envt, int save)
-{
-  jbuf jb, *old_jb;
-  
-  /* Initialize */
-  old_jb      = top_jmp_buf;
-  top_jmp_buf = &jb;
-
-  pc          = code;
-  constants   = consts;
-  env         = envt;
-
-  if (MY_SETJMP(jb)) {
-    /* We come back from an error. */
-    set_signal_mask(jb.blocked);
-  }
-
-  /* Jump in the VM */
-  run();
-  
-  /* Reset the jumbuf we have on entrance */
-  top_jmp_buf = old_jb;
-}
+////static void run_vm(STk_instr *code, SCM *consts, SCM envt, int save)
+////{
+////  jbuf jb, *old_jb;
+////  
+////  /* Initialize */
+////  old_jb      = top_jmp_buf;
+////  top_jmp_buf = &jb;
+////
+////  pc          = code;
+////  constants   = consts;
+////  env         = envt;
+////
+////  if (MY_SETJMP(jb)) {
+////    /* We come back from an error. */
+////    set_signal_mask(jb.blocked);
+////  }
+////
+////  /* Jump in the VM */
+////  run();
+////  
+////  /* Reset the jumbuf we have on entrance */
+////  top_jmp_buf = old_jb;
+////}
 
 
 void STk_execute_current_handler(SCM kind, SCM location, SCM message)
@@ -1469,8 +1546,7 @@ SCM STk_load_bcode_file(SCM f)
     }
 
     code = read_code(f, size);			    	   /* Read the code */
-
-    run_vm(code, VECTOR_DATA(consts), STk_current_module, 1); /* Execute code read */
+    run_vm(code, VECTOR_DATA(consts), STk_current_module); /* Execute code read */
   }
   
   /* restore machine state */
